"""
Test the constraints
"""
import unittest

from nose.tools import eq_

from cis.test.util import mock
from cis.data_io.hyperpoint import HyperPoint


class TestSepConstraint(unittest.TestCase):

    def test_all_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(lat=0.0, lon=0.0, alt=50.0, pres=50.0, t=dt.datetime(1984, 8, 29))

        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        # in each direction
        h_sep = 1000
        # 15m altitude seperation
        a_sep = 15
        # 1 day (and a little bit) time seperation
        t_sep = 'P1DT1M'
        # Pressure constraint is 50/40 < p_sep < 60/50
        p_sep = 1.22

        constraint = SepConstraint(h_sep=h_sep, a_sep=a_sep, p_sep=p_sep, t_sep=t_sep)

        # This should leave us with 9 points: [[ 22, 23, 24]
        #                                      [ 27, 28, 29]
        #                                      [ 32, 33, 34]]
        ref_vals = np.array([27., 28., 29., 32., 33., 34.])

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    def test_alt_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(lat=0.0, lon=0.0, alt=50.0, t=dt.datetime(1984, 8, 29))

        # 15m altitude seperation
        a_sep = 15

        constraint = SepConstraint(a_sep=a_sep)

        # This should leave us with 15 points:  [ 21.  22.  23.  24.  25.]
        # [ 26.  27.  28.  29.  30.]
        #                                       [ 31.  32.  33.  34.  35.]
        ref_vals = np.array([21., 22., 23., 24., 25., 26., 27., 28., 29., 30., 31., 32., 33., 34., 35.])

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    def test_horizontal_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(lat=0.0, lon=0.0, alt=50.0, t=dt.datetime(1984, 8, 29))

        # One degree near 0, 0 is about 110km in latitude and longitude, so 300km should keep us to within 3 degrees
        # in each direction
        constraint = SepConstraint(h_sep=1000)

        # This should leave us with 30 points
        ref_vals = np.reshape(np.arange(50) + 1.0, (10, 5))[:, 1:4].flatten()

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    def test_time_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(lat=0.0, lon=0.0, alt=50.0, t=dt.datetime(1984, 8, 29))

        # 1 day (and a little bit) time seperation
        constraint = SepConstraint(t_sep='P1dT1M')

        # This should leave us with 30 points
        ref_vals = np.reshape(np.arange(50) + 1.0, (10, 5))[:, 1:4].flatten()

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())

    def test_pressure_constraint_in_4d(self):
        from cis.collocation.col_implementations import SepConstraint
        import datetime as dt
        import numpy as np

        ug_data = mock.make_regular_4d_ungridded_data()
        ug_data_points = ug_data.get_non_masked_points()
        sample_point = HyperPoint(0.0, 0.0, 50.0, 24.0, dt.datetime(1984, 8, 29))

        constraint = SepConstraint(p_sep=2)

        # This should leave us with 20 points:  [  6.   7.   8.   9.  10.]
        # [ 11.  12.  13.  14.  15.]
        #                                       [ 16.  17.  18.  19.  20.]
        #                                       [ 21.  22.  23.  24.  25.]
        ref_vals = np.array([6., 7., 8., 9., 10., 11., 12., 13., 14., 15., 16., 17., 18., 19., 20., 21., 22., 23.,
                             24., 25.])

        new_points = constraint.constrain_points(sample_point, ug_data_points)
        new_vals = new_points.vals

        eq_(ref_vals.size, new_vals.size)
        assert (np.equal(ref_vals, new_vals).all())


if __name__ == '__main__':
    unittest.main()



"""
Plot a Taylor diagram comparing datasets. This is quite different to the other plot types and so does more formatting
itself than we normally would.

From Zak Kipling's implemention in https://scm.physics.ox.ac.uk/svn/climproc_python/trunk/scripts/taylor.py
"""
from itertools import cycle, repeat
from six.moves import zip
import logging
from matplotlib.transforms import Transform
import matplotlib.cm

import numpy as np
from cis.plotting.genericplot import APlot


class CosTransform(Transform):
    """Cosine transform for a coordinate axis"""
    input_dims = 1
    output_dims = 1
    is_separable = True
    has_inverse = True

    def transform_non_affine(self, a):
        return np.ma.cos(a)

    def inverted(self):
        return ArcCosTransform()


class ArcCosTransform(Transform):
    """Inverse cosine transform for a coordinate axis"""
    input_dims = 1
    output_dims = 1
    is_separable = True
    has_inverse = True

    def transform_non_affine(self, a):
        return np.ma.arccos(a)

    def inverted(self):
        return CosTransform()


class Bunch(object):
    """
    Bunch a dict up into a 'namespace'

    Courtesy of http://stackoverflow.com/questions/2597278/python-load-variables-in-a-dict-into-namespace
    """
    def __init__(self, adict):
        self.__dict__.update(adict)


class Taylor(APlot):

    def __init__(self, packed_data, labels=None, colors=None, markers=None, extend=0.0, fold=False, logv=False,
                 gammamax=None, solid=False, stdbiasmax=None, bias=None,
                 cbarscale=None, cbarorient=None, cbarlabel=None, *args, **kwargs):
        """
        A routine for producing taylor diagrams using matplotlib

        :param list packed_data: A list of CommonData objects to compare, the first item is taken as the reference
        :param list labels: List of labels for each of the datasets
        :param list colors: Specify colour(s) to use
        :param list markers: Specify marker(s) to use
        :param float extend: Extend plot for negative correlation, default 0.0
        :param bool fold: Fold plot for negative correlation or large variance, default False
        :param bool logv: Work on logarithms of data sets
        :param float gammamax: Fix maximum extent of radial axis
        :param bool solid: Use solid markers, default False
        :param float stdbiasmax: Fix maximum standardised bias
        :param string bias: Indicate bias using the specified method (colo[u]r, size or flag). Default is None
        :param float cbarscale: A scale factor for the colorbar
        :param string cbarorient: The colorbar orientation ('vertical' or 'horizontal')
        :param string cbarlabel: A label for the colorbar
        :param kwargs: Any other arguments for plot, such as alpha
        """
        # xaxis and yaxis are None in this special case
        super(Taylor, self).__init__(packed_data, None, None, *args, **kwargs)

        self.labels = [l or d.name() for l, d in zip(labels, packed_data)]
        self.colors = colors
        self.extend = extend
        self.fold = fold
        self.gammamax = gammamax
        self.solid = solid
        self.usrstdbiasmax = stdbiasmax
        self.bias = bias
        self.cbarscale = cbarscale
        self.cbarorient = cbarorient or 'horizontal'
        self.cbarlabel = cbarlabel

        self.itemwidth = self.itemwidth or 6

        if self.bias in ('color', 'colour'):
            self.solid = True
            if colors:
                raise ValueError('Cannot use colo[u]r kwarg with bias=colo[u]r')
        elif self.bias == 'size':
            if self.itemwidth:
                ValueError('Cannot use markersize kwarg with --bias=size')
        elif self.bias is not None and self.bias != 'flag':
            raise ValueError('Invalid bias argument specified, should be either colo[u]r, size or flag')

        if logv:
            packed_data = [np.ma.log(d.data) for d in packed_data]

        means = np.ma.array([np.ma.mean(d.data) for d in packed_data])
        sigmas = np.ma.array([np.ma.std(d.data, ddof=1) for d in packed_data])

        logging.info("Means: {}".format(means))
        # Treat the first array as the reference
        ref_mean = means[0]
        ref_sigma = sigmas[0]

        self.stdbiases = (means - ref_mean) / ref_sigma
        self.gammas = sigmas / ref_sigma

        # Calculate the correlations against the first array
        self.corrs = np.ma.array([np.ma.corrcoef(d.data, packed_data[0].data)[0, 1] for d in packed_data])
        self.corrs = np.ma.minimum(np.ma.maximum(self.corrs, -1.), 1.)

        if markers is None:
            if self.bias == 'size':
                #           -   0   +
                self.markers = ['v', 's', '^']
            else:
                if solid:
                    self.markers = ['o', 'v', '^', '<', '>', 's', 'p', 'h', 'D', '*']
                else:
                    self.markers = ['x', '+', 'o', 'v', '^', '<', '>', 's', 'p', 'h', 'D', '*']
        else:
            self.markers = markers

        mask = np.ma.getmaskarray(self.gammas)
        if self.gammamax and not fold:
            mask = np.logical_or(mask, self.gammas > self.gammamax)
        if self.bias:
            mask = np.logical_or(mask, np.ma.getmaskarray(self.stdbiases))
            if self.usrstdbiasmax:
                mask = np.ma.logical_or(mask, np.ma.abs(self.stdbiases) > self.usrstdbiasmax)

        self.gammamax = self.gammamax or np.ma.amax(np.ma.masked_where(mask, self.gammas))
        if self.bias:
            self.maxstdbias = self.usrstdbiasmax or np.ma.amax(np.ma.abs(np.ma.masked_where(mask, self.stdbiases)))
            self.minstdbias = min(np.ma.amin(np.ma.abs(np.ma.masked_where(mask[1:], self.stdbiases[1:]))), self.maxstdbias / 8)
            self.stdbiasrange = self.maxstdbias - self.minstdbias

        if self.bias:
            self.any_overbias = False
            self.any_underbias = False
        if extend > -1.:
            self.labels = [l + ' $-$' if c < extend else l for l, c in zip(self.labels, self.corrs)]
            if fold:
                if extend == 0.:
                    self.corrs = np.ma.abs(self.corrs)
                else:
                    self.corrs = np.ma.maximum(self.corrs, extend)
        if self.gammamax:
            self.labels = [l + ' $>$' if g > self.gammamax else l for l, g in zip(self.labels, self.gammas)]
            if fold:
                self.gammas = np.ma.minimum(self.gammas, self.gammamax)
        if self.bias and self.usrstdbiasmax:
            self.any_overbias = np.ma.any(self.stdbiases > self.usrstdbiasmax)
            self.any_underbias = np.ma.any(self.stdbiases < -self.usrstdbiasmax)
            self.labels = [l + ' B-' if b < -self.usrstdbiasmax else l for l, b in zip(self.labels, self.stdbiases)]
            self.labels = [l + ' B+' if b > self.usrstdbiasmax else l for l, b in zip(self.labels, self.stdbiases)]
            if fold:
                self.stdbiases = np.ma.maximum(np.ma.minimum(self.stdbiases, self.usrstdbiasmax), -self.usrstdbiasmax)

        if self.bias in ('color', 'colour'):
            self.mplkwargs['cmap'] = matplotlib.cm.get_cmap('RdBu_r')
            self.mplkwargs['cmap'].set_bad('gray')
            self.mplkwargs['norm'] = matplotlib.colors.Normalize(-self.maxstdbias, self.maxstdbias)

    def __call__(self, ax):
        from matplotlib.projections import PolarAxes
        from matplotlib.transforms import IdentityTransform, blended_transform_factory

        tr = blended_transform_factory(ArcCosTransform(), IdentityTransform()) + PolarAxes.PolarTransform()

        ax.axis["left"].set_axis_direction("left")  # point left ticks inwards
        ax.axis["left"].toggle(ticklabels=False, label=False)  # don't label left axis
        ax.axis["right"].toggle(ticklabels=True, label=True)  # label bottom axis instead
        ax.axis["right"].set_axis_direction("bottom")  # orient labels sensibly

        ax.axis["bottom"].set_visible(False)  # kill axis squeezed into bottom-left corner
        ax.axis["top"].toggle(ticklabels=True, label=True)  # label curved axis

        ax.axis["right"].label.set_text("$\sigma / \sigma_0$")
        ax.axis["top"].label.set_text("Correlation")

        ax.grid(True)

        ax2 = ax.get_aux_axes(tr)

        if self.bias != 'size':
            size = self.itemwidth

        logging.info("Label     Correlation     Gamma     StdBias     Marker     Size")
        for label, corr, gamma, stdbias, marker, color in zip(self.labels, self.corrs, self.gammas, self.stdbiases,
                                                              cycle(self.markers),
                                                              (cycle(self.colors) if self.colors
                                                               else repeat(None))):
            if self.bias == 'size':
                marker = self.markers[((stdbias > 0.) - (stdbias < 0.)) + 1]
                size = min(max((abs(stdbias) - self.minstdbias) / self.stdbiasrange, 0.), 1.)
                size = self.itemwidth * (0.25 + 1.75 * size)

            logging.info("        ".join([str(s) for s in (label, corr, gamma, stdbias, marker, size)]))

            if self.bias in ('color', 'colour'):
                if size is not None:
                    self.mplkwargs['s'] = size ** 2
                p = ax2.scatter(corr, gamma, c=stdbias, marker=marker, label=label, **self.mplkwargs)
                # So that correct colours are used in legend...
                p.update_scalarmappable()
            else:
                if size is not None:
                    self.mplkwargs['markersize'] = size
                if color is not None:
                    if ':' in color:
                        facecolor, edgecolor = [x or 'none' for x in color.split(':', 2)]
                        maincolor = facecolor if facecolor != 'none' else edgecolor
                        have_edge = (edgecolor not in 'none')
                        if not have_edge: edgecolor = facecolor
                    else:
                        maincolor = color
                        if self.solid:
                            facecolor, edgecolor = color, 'black'
                        else:
                            facecolor, edgecolor = 'none', color
                        have_edge = not self.solid
                    self.mplkwargs['markerfacecolor'] = facecolor
                    self.mplkwargs['markeredgecolor'] = edgecolor
                else:
                    have_edge = not self.solid

                for p in ax2.plot(corr, gamma, marker=marker,
                                  label=label, linestyle=' ', **self.mplkwargs):
                    if color is None:
                        maincolor = p.get_markerfacecolor()
                        if self.solid:
                            facecolor, edgecolor = maincolor, maincolor
                        else:
                            facecolor, edgecolor = 'none', maincolor

                            p.set_markeredgecolor(edgecolor)
                            p.set_markerfacecolor(facecolor)

                    if self.bias == 'flag':
                        std_cent_rmse = np.ma.sqrt(1. + gamma * (gamma - 2 * corr))
                        if std_cent_rmse > .0001:
                            cos_theta = corr
                            sin_theta = np.ma.sqrt(1 - corr ** 2)
                            cos_flag_angle = gamma * np.sqrt(1. - corr ** 2) / std_cent_rmse
                            sin_flag_angle = (corr * gamma - 1.) / std_cent_rmse
                            assert (np.abs(np.log(cos_flag_angle ** 2 + sin_flag_angle ** 2)) < .001)
                            ax.plot((gamma * cos_theta, gamma * cos_theta + stdbias * cos_flag_angle),
                                    (gamma * sin_theta, gamma * sin_theta - stdbias * sin_flag_angle),
                                    linestyle=('dashed' if not self.solid and facecolor == 'none' else 'solid'),
                                    marker=' ', color=maincolor)

        arc = np.linspace(0., np.pi, 50)
        for radius in np.arange(0., 2 * self.gammamax, .2):
            ax.plot(1.0 + np.cos(arc) * radius, np.sin(arc) * radius, linestyle='dotted', color='green')

        ax2.plot(np.linspace(1.0, self.extend, 50),
                 np.linspace(1.0, 1.0, 50),
                 linestyle='dashed', color='black')

        if self.extend < 0.:
            ax.plot((0., 0.), ax.get_ylim(), linestyle='solid', color='black')

        ax.legend(loc='upper right', numpoints=1, scatterpoints=1)

        if self.bias in ('color', 'colour'):
            extend = 'neither'
            if self.fold and self.usrstdbiasmax:
                if self.any_overbias and self.any_underbias:
                    extend = 'both'
                elif self.any_overbias:
                    extend = 'max'
                elif self.any_underbias:
                    extend = 'min'
            if self.cbarscale:
                cbar_kwargs = dict(shrink=self.cbarscale)
            else:
                cbar_kwargs = dict(fraction=0.046, pad=0.1)
            cb = ax.get_figure().colorbar(p, ax=ax, orientation=self.cbarorient, extend=extend, **cbar_kwargs)
            cb.set_label(self.cbarlabel or 'bias / $\sigma_0$')

        return ax

    def is_map(self):
        return False

    def set_log_scales(self, ax, logx, logy):
        pass  # Don't!



import logging
import os

import numpy as np

from cis.data_io.Coord import CoordList
from cis.exceptions import InvalidVariableError, FileFormatError
from cis.data_io.products import AProduct
from cis.data_io.ungridded_data import UngriddedCoordinates, UngriddedData, Metadata
from cis.utils import add_to_list_if_not_none, dimensions_compatible, listify
from cis.data_io.netcdf import get_metadata, get_netcdf_file_attributes, read_many_files_individually, \
    get_netcdf_file_variables


class NCAR_NetCDF_RAF_variable_name_selector(object):
    """
    NCAR-RAF file definition: for the purposes of this data product
        It is a NetCDF file
        It contains attributes and variables
        The attributes must contain the following:
            an attribute GASSP_Version
            or an attribute Conventions with the value "NCAR-RAF/nimbus"

    This should successful differentiate the names of the variables for the various dimensions in NCAR-RAF and GASSP.
    The variables set are:
    station: True if this is for a fixed station data, False otherwise
    station_latitude: the latitude of the station (set only if station is true)
    station_longitude: the longitude of the station (set only if station is true)
    altitude: Altitude of the measurement if it is fixed, none if this is not fixed, defaults to DEFAULT_ALTITUDE
    time_variable_name: variable name for time
    latitude_variable_name: variable name of the latitude, None if fixed
    longitude_variable_name: variable name of the longitude, None if fixed
    altitude_variable_name: variable name of the altitude, None if fixed
    pressure_variable_name: variable name of the pressure, None if not included in data
    time_stamp_info: time stamp info if included, None otherwise

    The choice are in order of preference:
    Aeroplane GASSP (Data with time, lat, lon and altitude vars)
    Ship GASSP (Data with time, lat and lon vars but fixed altitude)
    Station GASSP (Data with time vars but fixed lat, lon and altitude)
    NCAR-RAF: (Data with time, lat, lon and altitude, but names are defined in coordinates attribute)
    """

    # Attribute name for the attribute holding the time coordinate
    TIME_COORDINATE_NAME = 'Time_Coordinate'
    # Attribute name for the attribute holding the latitude coordinate
    LATITUDE_COORDINATE_NAME = 'Latitude_Coordinate'
    # Attribute name for the attribute holding the longitude coordinate
    LONGITUDE_COORDINATE_NAME = 'Longitude_Coordinate'
    # Attribute name for the attribute holding the altitude or vertical coordinate
    ALTITUDE_COORDINATE_NAME = 'Vertical_Coordinate'
    # Variable name for the dataset for corrected pressure
    CORRECTED_PRESSURE_VAR_NAME = 'PSXC'
    # Variable name for the dataset for pressure
    PRESSURE_VAR_NAME = 'PRE'
    # Attribute name for the attribute holding the names of the best coordinates to use for lon, lat, alt and time
    BEST_COORDINATES_NAME = 'Coordinates'
    # Attribute name for the attribute holding the stations latitude
    STATION_LATITUDE_NAME = 'Station_Lat'
    # Attribute name for the attribute holding the stations longitude
    STATION_LONGITUDE_NAME = 'Station_Lon'
    # Attribute name for the attribute holding the stations altitude
    STATION_ALTITUDE_NAME = 'Station_Altitude'
    # Attribute name for the attribute holding the time stamp info (this is used for where time is measured from
    TIME_STAMP_INFO_NAME = 'Time_Stamp_Info'
    # The default altitude to use if it is not set, for instance ship based GASSP does not contain altitude
    DEFAULT_ALTITUDE = 0

    def __init__(self, attributes, variables):
        """
        Initialisation
        :param attributes: dictionary of attributes and their values (or list of dictionarys if multiple files read)
        :param variables: dictionary of variable names and NetCDF Variable objects
        (or list of dictionarys if multiple files read)
        :return: nothing
        """
        self.station = False
        self.station_latitude = None
        self.latitude_variable_name = None
        self.station_longitude = None
        self.longitude_variable_name = None
        self.altitude = None
        self.altitude_variable_name = None
        self.pressure_variable_name = None
        self.time_stamp_info = None
        self.time_dimensions = None

        self._attributes = [{k.lower(): v for k, v in list(attrs.items())} for attrs in listify(attributes)]
        if len(variables) == 0:
            raise InvalidVariableError("No variables in the file so the type of data is unknown")
        self._variables = list(variables[0].keys())
        self._variable_dimensions = [{name: var.dimensions for name, var in list(vars.items())}
                                     for vars in listify(variables)]
        self._check_has_variables_and_attributes()

        # Carry out these checks using the attributes from the first file as a 'master'
        if self.TIME_COORDINATE_NAME.lower() in self._attributes[0]:
            self.time_variable_name = self._get_coordinate_variable_name(self.TIME_COORDINATE_NAME, "time")

            if self.LATITUDE_COORDINATE_NAME.lower() in self._attributes[0]:
                self._lat_lon_var_specified_setup()
            elif self.STATION_LATITUDE_NAME.lower() in self._attributes[0]:
                self._stationary_setup()
            else:
                raise InvalidVariableError("No attributes indicating latitude, expecting '{}' or '{}'"
                                           .format(self.STATION_LATITUDE_NAME, self.LONGITUDE_COORDINATE_NAME))
        elif self.BEST_COORDINATES_NAME.lower() in self._attributes[0]:
            self._best_coordinates_setup()
        else:
            raise InvalidVariableError(
                "No attributes indicating time variable name, expecting either '{}' or 'Coordinates'"
                .format(self.TIME_COORDINATE_NAME))

        if self.CORRECTED_PRESSURE_VAR_NAME in self._variables:
            self.pressure_variable_name = self.CORRECTED_PRESSURE_VAR_NAME
        elif self.PRESSURE_VAR_NAME in self._variables:
            self.pressure_variable_name = self.PRESSURE_VAR_NAME
        else:
            self.pressure_variable_name = None

        if self.TIME_STAMP_INFO_NAME.lower() in self._attributes[0]:
            # Not all files will have the same timestamp -> Retrieve a list of timestamps for each file.
            self.time_stamp_info = [attrs[self.TIME_STAMP_INFO_NAME.lower()] for attrs in self._attributes]

        self.time_dimensions = self._variable_dimensions[0][self.time_variable_name]

    def _get_coordinate_variable_name(self, attribute_name, coordinate_display_name):
        """
        Reads an attribute value for a co-ordinate and returns the value. Checks that value is a variables in the data
        Throws InvalidVariableError if the attribute or variable does not exist
        :param attribute_name: the name of the attribute to read
        :param coordinate_display_name: the display name of the attribute read
        :return: the variable name
        """
        if attribute_name.lower() in self._attributes[0]:
            variable_name = self._attributes[0][attribute_name.lower()]
            if variable_name not in self._variables:  # Just check the first file
                raise InvalidVariableError("There is no variable for the {} co-ordinate '{}'"
                                           .format(coordinate_display_name, variable_name))
            return variable_name

        raise InvalidVariableError(
            "No attributes indicating {} variable name, expecting '{}'"
            .format(coordinate_display_name, attribute_name))

    def _parse_station_altitude(self, altitude_string):
        """
        Parse a station's altitude string. Will try and read it directly as a float, then will try and read the first
        white-space separated part of the string (e.g. '80 m asl' -> float(80)), finally sets the default altitude and
        logs a warning indicating this.
        :param altitude_string:
        :return:
        """
        try:
            return float(altitude_string)
        except ValueError:
            try:
                return float(altitude_string.split()[0])
            except ValueError:
                logging.warning("Couldn't parse station altitude attribute '{alt_string}'. Using default altitude "
                                "({default}m)".format(alt_string=altitude_string, default=self.DEFAULT_ALTITUDE))
                return self.DEFAULT_ALTITUDE

    @staticmethod
    def _parse_station_lat_lon(lat_lon_string):
        """
        Parse a station's latitude or longitude string. Will try and read it directly as a float, otherwise will try and
         read the first white-space separated part of the string (e.g. '80 degrees north' -> float(80)).
        :param lat_lon_string:
        :return:
        """
        from cis.exceptions import InvalidVariableError
        try:
            return float(lat_lon_string)
        except ValueError:
            try:
                return float(lat_lon_string.split()[0])
            except ValueError:
                raise InvalidVariableError("Couldn't parse station attribute '{}'".format(lat_lon_string))

    def _stationary_setup(self):
        """
        Set up object when latitude and longitude are fixed
        """
        from cis.exceptions import InvalidVariableError
        if self.STATION_LATITUDE_NAME.lower() not in self._attributes[0]:
            raise InvalidVariableError("No attributes indicating latitude, expecting '{}'"
                                       .format(self.STATION_LATITUDE_NAME))
        # We need a bunch of different latitudes for different files
        self.station_latitude = [self._parse_station_lat_lon(attr[self.STATION_LATITUDE_NAME.lower()])
                                 for attr in self._attributes]

        if self.STATION_LONGITUDE_NAME.lower() not in self._attributes[0]:
            raise InvalidVariableError("No attributes indicating longitude, expecting '{}'"
                                       .format(self.STATION_LONGITUDE_NAME))
        self.station_longitude = [self._parse_station_lat_lon(attr[self.STATION_LONGITUDE_NAME.lower()])
                                  for attr in self._attributes]
        self.station = True

        if self.STATION_ALTITUDE_NAME.lower() in self._attributes[0]:
            self.altitude = [self._parse_station_altitude(attr[self.STATION_ALTITUDE_NAME.lower()])
                             for attr in self._attributes]
        else:
            self.altitude = [self.DEFAULT_ALTITUDE for attr in self._attributes]

    def _check_has_variables_and_attributes(self):
        """
        Check that netcdf file has variables and attributes
        """
        if self._variables is None or len(self._variables) == 0:
            raise InvalidVariableError("No variables in the file so the type of data is unknown")
        if self._attributes[0] is None or len(self._attributes[0]) == 0:
            raise InvalidVariableError("No attributes in the file so type of data is unknown")

    def _lat_lon_var_specified_setup(self):
        """
        Set up object when latitude and longitude variables are being used
        """

        self.latitude_variable_name = self._get_coordinate_variable_name(self.LATITUDE_COORDINATE_NAME, "latitude")
        self.longitude_variable_name = self._get_coordinate_variable_name(self.LONGITUDE_COORDINATE_NAME, "longitude")

        if self.ALTITUDE_COORDINATE_NAME.lower() in self._attributes[0]:
            self.altitude_variable_name = self._get_coordinate_variable_name(self.ALTITUDE_COORDINATE_NAME, "vertical")
        else:
            self.altitude = [self.DEFAULT_ALTITUDE for attr in self._attributes]

    def _best_coordinates_setup(self):
        """
        Set up object when coordinates attribute is found
        """
        coordinates_vars = []
        coordinates = self._attributes[0][self.BEST_COORDINATES_NAME.lower()]
        if coordinates is not None:
            coordinates_vars = coordinates.split()  # split on whitespace

        if len(coordinates_vars) is not 4:
            raise InvalidVariableError('The coordinate attribute does not have four entries. '
                                       'It should be space separated "longitude latitude altitude time"')

        for coordinates_var in coordinates_vars:
            if coordinates_var not in self._variables:
                raise InvalidVariableError("There is no variable for the co-ordinate '{}'".format(coordinates_var))

        self.longitude_variable_name, \
            self.latitude_variable_name, \
            self.altitude_variable_name, \
            self.time_variable_name \
            = coordinates_vars

    def find_auxiliary_coordinate(self, variable):
        """
        Find the variable name of an auxiliary coordinate for the given variable (if there is one).

        :param str variable: The data variable we're checking for any auxiliary coordinates
        :return str or None: The name of the variable holding the auxiliary coordinate or None
        """
        aux_coord_name = None
        dim_coord_names = [self.latitude_variable_name, self.longitude_variable_name,
                           self.altitude_variable_name, self.pressure_variable_name] + list(self.time_dimensions)
        # Find the *dimension* which corresponds to the auxiliary coordinate
        aux_coords = [dim for dim in self._variable_dimensions[0][variable] if dim not in dim_coord_names]
        if len(aux_coords) > 1:
            raise InvalidVariableError("CIS currently only supports reading data variables with one auxilliary "
                                       "coordinate")
        elif len(aux_coords) == 1:
            # If there is also a variable named after that dimension then this is the variable we're after
            if aux_coords[0] in self._variable_dimensions[0]:
                aux_coord_name = aux_coords[0]
            # Otherwise we need to look through all the variables and choose the first variable whose dimension is only
            #  the auxiliary dimension.
            else:
                for v, dims in self._variable_dimensions[0].items():
                    if dims[0] == aux_coords[0]:
                        aux_coord_name = v
                        break
        return aux_coord_name

    def get_variable_names_which_have_time_coord(self):
        variables = []
        for name, dimensions in list(self._variable_dimensions[0].items()):
            if len(dimensions) > 0 and dimensions_compatible(dimensions, self.time_dimensions):
                variables.append(name)
        return set(variables)


class NCAR_NetCDF_RAF(AProduct):
    """
    Data product for NCAR-RAF NetCDF files. This includes the subset of GASSP (which is its major use case)
    """

    # Set the priority to be higher than the other netcdf file types
    priority = 20

    # Name of the attribute in GASSP that defines the GASSP version
    GASSP_VERSION_ATTRIBUTE_NAME = "GASSP_Version"

    # Name of the attribute in NCAR_RAF that defines the converntion
    NCAR_RAF_CONVENTIONS_ATTRIBUTE_NAME = "Conventions"
    # Known NCAR RAF Converntions
    NCAR_RAF_KNOWN_CONVENTION = "NCAR-RAF/nimbus"
    # NCAR RAF Convention version name
    NCAR_RAF_CONVENTION_VERSION_ATTRIBUTE_NAME = "ConventionsVersion"

    def __init__(self, variable_selector_class=NCAR_NetCDF_RAF_variable_name_selector):
        """
        Setup NCAR RAF data product, allow a different variable selector class if needed
        :param variable_selector_class: the class to use for variable selection
        :return:nothing
        """
        super(NCAR_NetCDF_RAF, self).__init__()
        self.variableSelectorClass = variable_selector_class
        self.valid_dimensions = ["Time"]

    def get_file_signature(self):
        """
        Get the filename possibilities allowed for NCAR-RAF data
        :return: list of regular expressions for ncar raf data
        """
        return [r'.*\.nc$']

    def _get_file_format(self, filename):
        """
        Get the file type and any errors
        :param filename: the filename to read the type of
        :return: filetype
        :raises: FileFormatError if there is an error
        """
        if not os.path.isfile(filename):
            raise FileFormatError(["File does not exist"])
        try:
            attributes = get_netcdf_file_attributes(filename)
        except (RuntimeError, IOError) as ex:
            raise FileFormatError(["File is unreadable", ex.args[0]])

        attributes_lower = {attr.lower(): val for attr, val in list(attributes.items())}
        if self.GASSP_VERSION_ATTRIBUTE_NAME.lower() in attributes_lower:
            file_type = "NetCDF/GASSP/{}".format(attributes_lower[self.GASSP_VERSION_ATTRIBUTE_NAME.lower()])

        elif self.NCAR_RAF_CONVENTIONS_ATTRIBUTE_NAME.lower() in attributes_lower:
            ncarraf_convention = attributes_lower[self.NCAR_RAF_CONVENTIONS_ATTRIBUTE_NAME.lower()]
            version = ""
            if self.NCAR_RAF_CONVENTION_VERSION_ATTRIBUTE_NAME.lower() in attributes_lower:
                version = "/{}".format(attributes_lower[self.NCAR_RAF_CONVENTION_VERSION_ATTRIBUTE_NAME.lower()])
            if not ncarraf_convention == self.NCAR_RAF_KNOWN_CONVENTION:
                raise FileFormatError(["NCAR-RAF convention unknown, expecting '{}' was '{}'"
                                      .format(self.NCAR_RAF_KNOWN_CONVENTION, ncarraf_convention)])
            file_type = "NetCDF/{}{}".format(ncarraf_convention, version)

        else:
            raise FileFormatError(["File does not appear to be NCAR RAF or GASSP. No attributes for either '{}' or '{}'"
                                  .format(self.GASSP_VERSION_ATTRIBUTE_NAME, self.NCAR_RAF_CONVENTIONS_ATTRIBUTE_NAME)])

        return file_type

    def get_file_type_error(self, filename):
        """
        Test that the file is of the correct signature
        :param filename: the file name for the file
        :return: list fo errors or None
        """
        try:
            self._get_file_format(filename)
            return None
        except FileFormatError as ex:
            return ex.error_list

    def _load_data_definition(self, filenames):
        """
        Load the definition of the data
        :param filenames: filenames from which to load the data
        :return: variable selector containing the data definitions
        """
        variables_list = [get_netcdf_file_variables(f) for f in filenames]
        attributes = [get_netcdf_file_attributes(f) for f in filenames]

        variable_selector = self.variableSelectorClass(attributes, variables_list)
        return variable_selector

    def _load_data(self, filenames, variable):
        """
        Open the file and find the correct variables to load in
        :param filenames: the filenames to load
        :param variable: an extra variable to load
        :return: a list of load data and the variable selector used to load name it
        """

        variable_selector = self._load_data_definition(filenames)

        variables_list = [variable_selector.time_variable_name]
        add_to_list_if_not_none(variable_selector.latitude_variable_name, variables_list)
        add_to_list_if_not_none(variable_selector.longitude_variable_name, variables_list)
        add_to_list_if_not_none(variable_selector.altitude_variable_name, variables_list)
        add_to_list_if_not_none(variable_selector.pressure_variable_name, variables_list)

        logging.info("Listing coordinates: " + str(variables_list))
        add_to_list_if_not_none(variable, variables_list)

        data_variables = read_many_files_individually(filenames, variables_list)

        return data_variables, variable_selector

    def _create_coordinates_list(self, data_variables, variable_selector):
        """
        Create a co-ordinate list for the data
        :param data_variables: the load data
        :param variable_selector: the variable selector for the data
        :return: a list of coordinates
        """
        coords = CoordList()

        # Time
        time_coord = self._create_time_coord(variable_selector.time_stamp_info, variable_selector.time_variable_name,
                                             data_variables)
        coords.append(time_coord)

        # Lat and Lon
        # Multiple points counts for multiple files
        points_count = [np.product(var.shape) for var in data_variables[variable_selector.time_variable_name]]
        if variable_selector.station:
            lat_coord = self._create_fixed_value_coord("Y", variable_selector.station_latitude, "degrees_north",
                                                       points_count, "latitude")
            lon_coord = self._create_fixed_value_coord("X", variable_selector.station_longitude, "degrees_east",
                                                       points_count, "longitude")
        else:
            lat_coord = self._create_coord("Y", variable_selector.latitude_variable_name, data_variables, "latitude")
            lon_coord = self._create_coord("X", variable_selector.longitude_variable_name, data_variables, "longitude")
        coords.append(lat_coord)
        coords.append(lon_coord)

        # Altitude
        if variable_selector.altitude is None:
            altitude_coord = self._create_coord("Z", variable_selector.altitude_variable_name, data_variables,
                                                "altitude")
        else:
            altitude_coord = self._create_fixed_value_coord("Z", variable_selector.altitude, "meters", points_count,
                                                            "altitude")
        coords.append(altitude_coord)

        # Pressure
        if variable_selector.pressure_variable_name is not None:
            coords.append(
                self._create_coord("P", variable_selector.pressure_variable_name, data_variables, "air_pressure"))
        return coords

    @staticmethod
    def _add_aux_coordinate(dim_coords, filename, aux_coord_name, length):
        """
        Add an auxiliary coordinate to a list of (reshaped) dimension coordinates

        :param dim_coords: CoordList of one-dimensional coordinates representing physical dimensions
        :param filename: The data file containing the aux coord
        :param aux_coord_name: The name of the aux coord to add to the coord list
        :param length: The length of the data dimensions which this auxiliary coordinate should span
        :return: A CoordList of reshaped (2D) physical coordinates plus the 2D auxiliary coordinate
        """
        from cis.data_io.Coord import Coord
        from cis.utils import expand_1d_to_2d_array
        from cis.data_io.netcdf import read

        # We assume that the auxilliary coordinate is the same shape across files
        d = read(filename, [aux_coord_name])[aux_coord_name]
        # Reshape to the length given
        aux_data = expand_1d_to_2d_array(d[:], length, axis=0)
        # Get the length of the auxiliary coordinate
        len_y = d[:].size

        for dim_coord in dim_coords:
            dim_coord.data = expand_1d_to_2d_array(dim_coord.data, len_y, axis=1)

        all_coords = dim_coords + [Coord(aux_data, get_metadata(d))]

        return all_coords

    def create_coords(self, filenames, variable=None):
        """
        Reads the coordinates and data if required from the files
        :param filenames: List of filenames to read coordinates from
        :param variable: load a variable for the data
        :return: Coordinates
        """
        data_variables, variable_selector = self._load_data(filenames, variable)

        dim_coords = self._create_coordinates_list(data_variables, variable_selector)

        if variable is None:
            return UngriddedCoordinates(dim_coords)
        else:
            aux_coord_name = variable_selector.find_auxiliary_coordinate(variable)
            if aux_coord_name is not None:
                all_coords = self._add_aux_coordinate(dim_coords, filenames[0], aux_coord_name,
                                                      dim_coords.get_coord(standard_name='time').data.size)
            else:
                all_coords = dim_coords
            return UngriddedData(data_variables[variable], get_metadata(data_variables[variable][0]), all_coords)

    def create_data_object(self, filenames, variable):
        """
        Load the variable with it coordinates from the files
        :param filenames: filenames of the file
        :param variable: the variable to load
        :return: Data object
        """
        return self.create_coords(filenames, variable)

    def _create_coord(self, coord_axis, data_variable_name, data_variables, standard_name):
        """
        Create a coordinate for the co-ordinate list
        :param coord_axis: axis of the coordinate in the coords
        :param data_variable_name: the name of the variable in the data
        :param data_variables: the data variables
        :param standard_name: the standard name it should have
        :return: a coords object
        """
        from cis.data_io.Coord import Coord

        coordinate_data_objects = []
        for d in data_variables[data_variable_name]:
            m = get_metadata(d)
            m.standard_name = standard_name
            coordinate_data_objects.append(Coord(d, m, coord_axis))

        return Coord.from_many_coordinates(coordinate_data_objects)

    def _create_time_coord(self, timestamp, time_variable_name, data_variables, coord_axis='T', standard_name='time'):
        """
        Create a time coordinate, taking into account the fact that each file may have a different timestamp.
        :param timestamp: Timestamp or list of timestamps for
        :param time_variable_name: Name of the time variable
        :param data_variables: Dictionary containing one or multiple netCDF data variables for each variable name
        :param coord_axis: Axis, default 'T'
        :param standard_name: Coord standard name, default 'time'
        :return: Coordinate
        """
        from cis.data_io.Coord import Coord
        from six.moves import zip_longest

        timestamps = listify(timestamp)
        time_variables = data_variables[time_variable_name]
        time_coords = []
        # Create a coordinate for each separate file to account for differing timestamps
        for file_time_var, timestamp in zip_longest(time_variables, timestamps):
            metadata = get_metadata(file_time_var)
            metadata.standard_name = standard_name
            coord = Coord(file_time_var, metadata, coord_axis)
            coord.convert_to_std_time(timestamp)
            time_coords.append(coord)

        return Coord.from_many_coordinates(time_coords)

    def _create_fixed_value_coord(self, coord_axis, values, coord_units, points_counts, coord_name):
        """
        Create a coordinate with a fixed value
        :param coord_axis: Axis of the coordinate in the coords
        :param coord_name: The name of the coordinate
        :param coord_units: The units for the coordinate
        :param points_counts: Number of points for this coordinate, or list of sizes for multiple files
        :param values: Value of coordinate, or list of values for multiple files
        :return:
        """
        from cis.data_io.Coord import Coord

        values = listify(values)
        points_counts = listify(points_counts)
        all_points = np.array([])
        # Create separate arrays with values and sizes corresponding to each of the different input files.
        for value, points_count in zip(values, points_counts):
            file_points = np.ma.array(np.zeros(points_count) + float(value))
            all_points = np.append(all_points, file_points)
        metadata = Metadata(name=coord_name, shape=all_points.shape, units=coord_units,
                            range=(min(values), max(values)))
        return Coord(all_points, metadata, coord_axis)

    def get_variable_names(self, filenames, data_type=None):
        """
        Get a list of available variable names
        This can be overridden in specific products to improve on this
        """

        selector = self._load_data_definition(filenames)
        return selector.get_variable_names_which_have_time_coord()

    def get_file_format(self, filename):
        """
        Return the file format, in general this string is parent format/specific instance/version
        e.g. NetCDF/GASSP/1.0
        :param str filename: Filename of a an example file from the dataset
        :return: File format, of the form parent format/specific instance/version
        :rtype: str
        :raises: FileFormatError if files type is not determinable
        """
        return self._get_file_format(filename)



import unittest

from cis.collocation.col_framework import get_kernel
from cis.test.util import mock
from cis.aggregation.collapse_kernels import aggregation_kernels, CountKernel
from cis.test.utils_for_testing import *
from cis.data_io.gridded_data import make_from_cube


class TestMomentsKernel(unittest.TestCase):

    def test_GIVEN_gridded_data_WHEN_full_collapse_THEN_calculations_correct(self):
        cube = make_from_cube(mock.make_mock_cube())
        kernel = aggregation_kernels['moments']
        result = cube.collapsed(['y'], how=kernel)

        expected_means = numpy.array([7, 8, 9])
        expected_std_dev = numpy.array(3 * [numpy.sqrt(22.5)])
        expected_no = numpy.array([5, 5, 5])
        assert_that(len(result), is_(3))
        assert_that(numpy.allclose(result[0].data, expected_means))
        assert_that(numpy.allclose(result[1].data, expected_std_dev))
        assert_that(numpy.array_equal(result[2].data, expected_no))

    def test_GIVEN_gridded_data_WHEN_full_collapse_THEN_metadata_correct(self):
        cube = make_from_cube(mock.make_mock_cube())
        cube.standard_name = 'age_of_sea_ice'  # Use a CF compliant name
        cube.long_name = 'Age of sea ice'
        cube.var_name = 'age_ice'
        cube.units = 'years'
        kernel = aggregation_kernels['moments']
        result = cube.collapsed(['y'], how=kernel)

        mean, stddev, num = result
        assert_that(mean.standard_name, is_('age_of_sea_ice'))
        assert_that(stddev.standard_name, is_(None))
        assert_that(num.standard_name, is_(None))
        assert_that(mean.long_name, is_('Age of sea ice'))
        assert_that(stddev.long_name, is_('Corrected sample standard deviation of Age of sea ice'))
        assert_that(num.long_name, is_('Number of points used to calculate the mean of Age of sea ice'))
        assert_that(mean.var_name, is_('age_ice'))
        assert_that(stddev.var_name, is_('age_ice_std_dev'))
        assert_that(num.var_name, is_('age_ice_num_points'))
        assert_that(mean.units, is_('years'))
        assert_that(stddev.units, is_('years'))
        assert_that(num.units, is_(None))

    def test_GIVEN_grid_contains_single_points_WHEN_collapse_THEN_stddev_undefined(self):
        cube = make_from_cube(mock.make_mock_cube(2, 2))
        cube.data = numpy.ma.masked_invalid([[float('Nan'), 1], [float('Nan'), float('Nan')]])
        kernel = aggregation_kernels['moments']
        result = cube.collapsed(['y'], how=kernel)

        assert_that(result[1].data.mask.all())

    def test_GIVEN_ungridded_data_WHEN_collapse_THEN_calculations_correct(self):
        grid = {'y': slice(-12.5, 12.5, 12.5)}
        data = mock.make_regular_2d_ungridded_data()
        kernel_class = get_kernel('moments')
        kernel = kernel_class()
        result = data.aggregate(how=kernel, **grid)

        expected_means = numpy.array([3.5, 11])
        expected_std_dev = numpy.array([numpy.sqrt(3.5), numpy.sqrt(7.5)])
        expected_no = numpy.array([6, 9])
        assert_that(len(result), is_(3))
        assert_arrays_almost_equal(result[0].data.flatten(), expected_means)
        assert_arrays_almost_equal(result[1].data.flatten(), expected_std_dev)
        assert_that(numpy.array_equal(result[2].data.flatten(), expected_no))

    def test_GIVEN_ungridded_data_WHEN_collapse_THEN_metadata_correct(self):
        grid = {'y': slice(-10, 10, 10)}
        data = mock.make_regular_2d_ungridded_data()
        kernel_class = get_kernel('moments')
        kernel = kernel_class()
        result = data.aggregate(how=kernel, **grid)

        mean, stddev, num = result
        assert_that(mean.standard_name, is_('rainfall_rate'))
        assert_that(stddev.standard_name, is_(None))
        assert_that(num.standard_name, is_(None))
        assert_that(mean.long_name, is_('TOTAL RAINFALL RATE: LS+CONV KG/M2/S'))
        assert_that(stddev.long_name,
                    is_('Corrected sample standard deviation of TOTAL RAINFALL RATE: LS+CONV KG/M2/S'))
        assert_that(num.long_name, is_('Number of points used to calculate the mean of '
                                       'TOTAL RAINFALL RATE: LS+CONV KG/M2/S'))
        assert_that(mean.var_name, is_('rain'))
        assert_that(stddev.var_name, is_('rain_std_dev'))
        assert_that(num.var_name, is_('rain_num_points'))
        assert_that(mean.units, is_('kg m-2 s-1'))
        assert_that(stddev.units, is_('kg m-2 s-1'))
        assert_that(num.units, is_(None))


class TestCountKernel(unittest.TestCase):

    def test_GIVEN_missing_data_WHEN_count_THEN_calculation_correct(self):
        cube = mock.make_5x3_lon_lat_2d_cube_with_missing_data()
        kernel = CountKernel()
        data = kernel.count_kernel_func(cube.data, 0)
        assert_that(numpy.array_equal(data, numpy.array([4, 4, 4])))

if __name__ == '__main__':
    unittest.main()



import netCDF4

import numpy
from hamcrest import assert_that, is_
import unittest

from cis.cis_main import evaluate_cmd, col_cmd
from cis.test.integration.base_integration_test import BaseIntegrationTest
from cis.test.integration_test_data import *
from cis.parse import parse_args
from cis.test.unit.eval.test_calc import compare_masked_arrays


class TestEval(BaseIntegrationTest):

    def test_Aeronet_wavelength_calculation(self):
        # Example from the CIS Phase 3 Software spec:
        # ... a user should be able to write a plugin to calculate the Aeronet AOD at 550nm from the AOD at 500 nm as
        # AOD550 = AOD500 * (550/500)^(-1*Angstrom500-870)"
        # Takes 3s
        args = ['eval', 'AOT_500,500-870Angstrom=a550to870:' + escape_colons(another_valid_aeronet_filename),
                'AOT_500 * (550.0/500)**(-1*a550to870)', '1', '-o', self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        evaluate_cmd(arguments)

        # Check correct:
        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        calculated_result = self.ds.variables['calculated_variable'][:]
        expected_result = [0.2341039087, 0.2285401152, 0.2228799533, 0.1953746746, 0.2094051561, 0.1696889668,
                           0.3137791803, 0.2798929273, 0.1664194279, 0.1254619092, 0.1258309124, 0.1496960031,
                           0.0768447737, 0.0550896430, 0.0534543107, 0.0538315909, 0.0666742975, 0.0512935449,
                           0.0699585189, 0.0645033944]
        assert_that(calculated_result.shape, is_((3140,)))
        assert numpy.allclose(expected_result, calculated_result[0:20])

    def test_ECHAMHAM_wavelength_sum(self):
        args = ['eval', "%s,%s:%s" % (valid_echamham_variable_1, valid_echamham_variable_2, escape_colons(valid_echamham_filename)),
                '%s+%s' % (valid_echamham_variable_1, valid_echamham_variable_2), '1', '-o', self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        evaluate_cmd(arguments)

        # Check correct:
        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        calculated_result = self.ds.variables['calculated_variable'][:]
        # A hand calculated selection of values
        expected_result = [0.007633533, 0.007646653, 0.007749859, 0.007744226, 0.007761176]

        assert_that(calculated_result.shape, is_((96, 192)))
        assert numpy.allclose(expected_result, calculated_result[:][0][0:5])

    def test_collocated_NetCDF_Gridded_onto_GASSP(self):
        # First do a collocation of ECHAMHAM onto GASSP
        vars = valid_echamham_variable_1, valid_echamham_variable_2
        filename = escape_colons(valid_echamham_filename)
        sample_file = escape_colons(valid_GASSP_aeroplane_filename)
        sample_var = valid_GASSP_aeroplane_variable
        collocator_and_opts = 'nn[missing_data_for_missing_sample=True],variable=%s' % sample_var
        arguments = ['col', ",".join(vars) + ':' + filename,
                     sample_file + ':collocator=' + collocator_and_opts,
                     '-o', 'collocated_gassp']
        main_arguments = parse_args(arguments)
        col_cmd(main_arguments)

        # Check collocation is the same
        self.ds = netCDF4.Dataset('collocated_gassp.nc')
        col_var1 = self.ds.variables[valid_echamham_variable_1][:]
        col_var2 = self.ds.variables[valid_echamham_variable_2][:]
        # A hand calculated selection of values
        expected_col1 = numpy.ma.masked_invalid(
            [float('Nan'), float('Nan'), float('Nan'), 0.0814601778984, 0.0814601778984])
        compare_masked_arrays(expected_col1, col_var1[:][0:5])
        expected_col2 = numpy.ma.masked_invalid(
            [float('Nan'), float('Nan'), float('Nan'), 0.0741240680218, 0.0741240680218])
        compare_masked_arrays(expected_col2, col_var2[:][0:5])

        # Then do an evaluation using the collocated data:
        args = ['eval', "%s,%s:%s" % (valid_echamham_variable_1, valid_echamham_variable_2,
                                      'collocated_gassp.nc'),
                "%s=gassp_alias:%s" % (valid_GASSP_aeroplane_variable, escape_colons(valid_GASSP_aeroplane_filename)),
                "(%s + %s) / gassp_alias " % (valid_echamham_variable_1, valid_echamham_variable_2),
                '1', '-o', self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        evaluate_cmd(arguments)
        self.ds.close()

        # Check correct
        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        calculated_result = self.ds.variables['calculated_variable'][:]
        # A hand calculated selection of values
        expected_result = numpy.ma.masked_invalid([0.00196121983491, 0.00197255626472, 0.00120850731992])

        assert_that(calculated_result.shape, is_((311,)))
        # Check the first 3 vald values
        compare_masked_arrays(expected_result, calculated_result[:][10:13])
        os.remove('collocated_gassp.nc')

    @skip_pyhdf
    def test_CloudSat(self):
        args = ['eval', "%s,%s:%s" % (valid_cloudsat_RVOD_sdata_variable, valid_cloudsat_RVOD_vdata_variable,
                                      escape_colons(valid_cloudsat_RVOD_file)),
                '%s/%s' % (valid_cloudsat_RVOD_sdata_variable, valid_cloudsat_RVOD_vdata_variable), 'ppm', '-o',
                'cloudsat_var:' + self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        evaluate_cmd(arguments)
        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        assert_that(self.ds.variables['cloudsat_var'].units, is_('ppm'))

    def test_can_specify_output_variable(self):
        args = ['eval', "%s,%s:%s" % (valid_echamham_variable_1, valid_echamham_variable_2, escape_colons(valid_echamham_filename)),
                '%s+%s' % (valid_echamham_variable_1, valid_echamham_variable_2), 'kg m^-3',
                '-o', 'var_out:' + self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        evaluate_cmd(arguments)

        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        assert 'var_out' in self.ds.variables

    def test_can_specify_attributes_gridded(self):
        args = ['eval', "%s,%s:%s" % (valid_echamham_variable_1, valid_echamham_variable_2, escape_colons(valid_echamham_filename)),
                '%s+%s' % (valid_echamham_variable_1, valid_echamham_variable_2), 'kg m^-3',
                '-o', 'var_out:' + self.OUTPUT_FILENAME, '-a', 'att1=val1,att2=val2']
        arguments = parse_args(args)
        evaluate_cmd(arguments)

        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        assert_that(self.ds.variables['var_out'].att1, is_('val1'))
        assert_that(self.ds.variables['var_out'].att2, is_('val2'))

    def test_can_specify_units_gridded(self):
        args = ['eval', "%s,%s:%s" % (valid_echamham_variable_1, valid_echamham_variable_2, escape_colons(valid_echamham_filename)),
                '%s+%s' % (valid_echamham_variable_1, valid_echamham_variable_2), 'kg m^-3',
                '-o', 'var_out:' + self.OUTPUT_FILENAME, '-a', 'att1=val1,att2=val2']
        arguments = parse_args(args)
        evaluate_cmd(arguments)

        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        assert_that(self.ds.variables['var_out'].units, is_('kg m^-3'))

    def test_can_specify_units_gridded_no_output_var(self):
        args = ['eval', "%s:%s" % (valid_hadgem_variable, escape_colons(valid_hadgem_filename)), "od550aer", "ppm", "-o",
                self.OUTPUT_FILENAME, "-a", "att1=val1"]
        arguments = parse_args(args)
        evaluate_cmd(arguments)

        self.ds = netCDF4.Dataset(self.OUTPUT_FILENAME)
        assert_that(self.ds.variables['calculated_variable'].units, is_('ppm'))
        assert_that(self.ds.variables['calculated_variable'].att1, is_('val1'))



"""
Utilities for converting time units
"""
from cf_units import Unit

cis_standard_time_unit = Unit('days since 1600-01-01 00:00:00', calendar='gregorian')


class PartialDateTime(object):
    """
    A :class:`PartialDateTime` object specifies values for some subset of
    the calendar/time fields (year, month, hour, etc.) and defines a period of time,
    primarily for creating ranges of :class:`datetime.datetime` instances.
    """

    def __init__(self, year, month=None, day=None, hour=None, minute=None, second=None):
        """
        Allows creation of datetime ranges. The year is mandatory, all others are optional - but intermediate components
        must be specified. E.g. you can't specify a year and a day with no month.

        :param int year:
        :param int month:
        :param int day:
        :param int hour:
        :param int minute:
        :param int second:
        """
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second

        self.attributes = ['year', 'month', 'day', 'hour', 'minute', 'second']

        # Check we have valid components - there should be no non-None values after the first one (if there is one)
        components = [getattr(self, a) for a in self.attributes]
        if None in components and any(c is not None for c in components[components.index(None):]):
            raise ValueError("All intermediate values must be specified when creating a PartialDateTime")

        # Check that the components are valid by trying to find the minimum date
        try:
            _ = self.min()
        except TypeError:
            raise ValueError("Invalid PartialDateTime arguments")

    def range(self):
        """
        Return the datetime representing the start of the range defined by this object, and that representing the end.

        For example, to create a datetime range for the year 2008:

        >>> from cis.time_util import PartialDateTime
        >>> pdt = PartialDateTime(2008)
        >>> pdt.range()
        datetime(2008,1,1,1,1,1), datetime(2008,12,31,23,59,59)

        Or for February 2001:

        >>> pdt = PartialDateTime(2001, 2)
        >>> pdt.range()
        datetime(2001,2,1,1,1,1), datetime(2008,2,28,23,59,59)

        :return (datetime.datetime, datetime.datetime) : A pair of datetime objects
        """
        return self.min(), self.max()

    def min(self):
        """
        Return the datetime object for the start of the PartialDateTime period. I.e. the earliest valid datetime

        :return datetime.datetime: The earliest datetime
        """
        from datetime import datetime
        lower_limits = [None, 1, 1, 0, 0, 0]

        # Get the initialized date-components
        dt_start = [getattr(self, a) for a in self.attributes if getattr(self, a) is not None]

        # Fill in the remaining fields with default values.
        for idx in range(len(dt_start), len(self.attributes)):
            if self.attributes[idx] == 'day':
                dt_start.append(lower_limits[idx])
            else:
                dt_start.append(lower_limits[idx])
        return datetime(*dt_start)

    def max(self):
        """
        Return the datetime object for the end of the PartialDateTime period. I.e. the latest valid datetime

        :return datetime.datetime: The latest datetime
        """
        from datetime import datetime
        upper_limits = [None, 12, None, 23, 59, 59]

        # Get the initialized date-components
        dt_end = [getattr(self, a) for a in self.attributes if getattr(self, a) is not None]

        # Fill in the remaining fields with default values.
        for idx in range(len(dt_end), len(self.attributes)):
            if self.attributes[idx] == 'day':
                # Set day to last day of month. The year and month will already be in there.
                dt_end.append(find_last_day_of_month(*dt_end))
            else:
                dt_end.append(upper_limits[idx])

        return datetime(*dt_end)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


def find_last_day_of_month(year, month):
    """Finds the last day of a month.

    :param year: year as integer
    :param month: month as integer
    :return: last day of month as integer
    """
    import datetime
    next_month = datetime.datetime(year, month, 28) + datetime.timedelta(days=4)
    last_date = next_month - datetime.timedelta(days=next_month.day)
    return last_date.day


def calculate_mid_time(t1, t2):
    """
        Find the mid time between two times expressed as floats

    :param t1: a time represented as a float
    :param t2: a time in the same representation as t1
    :return: a float representing the time between t1 and t2
    """
    return t1 + (t2 - t1)/2.0


def convert_time_since_to_std_time(time_array, units):
    """
    Go via datetimes to be on the safe side.
    :param ndaray time_array:
    :param cf_units.Unit units:
    :return:
    """
    dt = units.num2date(time_array)
    return cis_standard_time_unit.date2num(dt)


def convert_time_using_time_stamp_info_to_std_time(time_array, units, time_stamp_info=None):
    """
    Convert the time using time stamp info and the first word of the units
    :param time_array: the time array to convert
    :param str units: the units of the array (e.g. day or Days from the file time reference 2012-12-12)
    :param str time_stamp_info: the time stamp to use for the convertion
    :return: converted data
    """
    from cf_units import Unit
    units = str(units).split()
    if len(units) is 0:
        raise ValueError("Units is empty when converting time")

    units_in_since_form = Unit(units[0] + " since " + time_stamp_info)

    return convert_time_since_to_std_time(time_array, units_in_since_form)


def convert_sec_since_to_std_time(seconds, ref):
    """
    Convert a number of seconds since a given reference datetime to a number of days since our standard time. The given
    reference DateTime must be on the Gregorian calendar.

    :param seconds: Array of seconds (since the reference time provided)
    :type: ndarray
    :param ref: The reference datetime which the seconds are counted from
    :type: DateTime
    :return: A numpy array containing all of the time values (in fractional days since the CIS standard time)
    """
    import numpy as np
    # Don't copy the array if this is a standard numpy array, unfortunately masked arrays don't have this option
    kwargs = {} if isinstance(seconds, np.ma.MaskedArray) else {'copy': False}
    days_since = seconds.astype('float64', **kwargs) / (3600*24.0)
    offset = ref - cis_standard_time_unit.num2date(0)
    return offset.days + days_since


def convert_std_time_to_datetime(std_time):
    return cis_standard_time_unit.num2date(std_time)


def convert_datetime_to_std_time(dt):
    return cis_standard_time_unit.date2num(dt)


def convert_julian_date_to_std_time(days_since):
    """
    Convert an array of julian days to cis standard time

    ..note:
        Array should have units like: Julian Date, days elapsed since 12:00 January 1, 4713 BC

    :param days_since: numpy array of fractional days since 12:00 January 1, 4713 BC
    :return: fractional days since cis standard time
    """
    from cf_units import date2julian_day
    offset = date2julian_day(cis_standard_time_unit.num2date(0), 'standard')
    return days_since - offset


def convert_cube_time_coord_to_standard_time(cube):
    """Converts the time coordinate from the one in the cube to one based on a standard time unit.
    :param cube: cube to modify
    :return: the cube
    """
    # Find the time coordinate.
    t_coord = cube.coord(standard_name='time')
    data_dim = cube.coord_dims(t_coord)
    if len(data_dim) > 0:
        # And remove it from the cube
        cube.remove_coord(t_coord)

        dt_points = t_coord.units.num2date(t_coord.points)
        new_datetime_nums = cis_standard_time_unit.date2num(dt_points)

        # new_datetime_nums = convert_numpy_array(t_coord.points, 'float64', convert_date)
        if t_coord.nbounds > 0:
            dt_bounds = t_coord.units.num2date(t_coord.bounds)
            new_bound_nums = cis_standard_time_unit.date2num(dt_bounds)
            t_coord.bounds = new_bound_nums

        # Create a new time coordinate by copying the old one, but using our new points and units
        new_time_coord = t_coord
        new_time_coord.points = new_datetime_nums
        new_time_coord.units = cis_standard_time_unit

        # And add the new coordinate back into the cube
        cube.add_dim_coord(new_time_coord, data_dim)

    return cube


def set_year(datetime, new_year):
    """
    Change the year of a datetime to some new year. If datetime is a leapday then return None
    :param datetime.datetime datetime: Datetime object to change
    :param int new_year: The new year
    :return: A datetime with the same date as the original except the changed year
    """
    # Once we have the data as datetimes we can just use replace to change the year...
    try:
        new_dt = datetime.replace(year=new_year)
    except ValueError:
        # ...Unless the date is 29th of Feb!
        new_dt = None
    return new_dt


def change_year_of_ungridded_data(data, new_year):
    """
     This slightly roundabout method works fine, but isn't particularly quick.
      I could just add the number of years times 365, but that does't take leap years into account. If I want to take
      leap years into account I can't use fractional days which would break the time. In principle I could take calculate
      the exact difference in integer days between the first date and the first date in the new year then apply that
      scaling - but THAT won't work if the data set spans a leap day...
    :param data: An ungridded data object to update in-place
    :param int new_year: The year to change the data to
    """
    import numpy as np

    dates = data.coord('time').data

    dt = convert_std_time_to_datetime(dates)

    np_set_year = np.vectorize(set_year)

    updated_dt = np_set_year(dt, new_year)
    new_dates = convert_datetime_to_std_time(updated_dt)

    data.coord('time').data = new_dates



from unittest import TestCase

from nose.tools import istest

from cis.data_io.gridded_data import GriddedDataList
from cis.data_io.ungridded_data import UngriddedDataList
from cis.collocation.col_implementations import mean, max, min, stddev, moments
from cis.test.utils_for_testing import *

from cis.test.util.mock import *


class TestUngriddedAggregation(TestCase):
    def setUp(self):
        self.kernel = mean()

    @istest
    def test_aggregating_single_point_in_one_dimension(self):
        grid = {'y': slice(-12.5, 12.5, 5)}

        data = make_dummy_ungridded_data_single_point()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[0], [0], [1.0], [0], [0]], mask=[[1], [1], [0], [1], [1]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_can_name_variables_by_standard_name(self):
        """
        Note that this is also the variable name for many ungridded cases
        """
        grid = {'latitude': slice(-12.5, 12.5, 5)}

        data = make_dummy_ungridded_data_single_point()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[0], [0], [1.0], [0], [0]], mask=[[1], [1], [0], [1], [1]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_aggregating_defaults(self):
        """
        Testing default aggregation edges. Note that the upper edges of the
        ungridded data points array are removed from the aggregate because they fall on the respective cells' upper
        bounds.
        """
        grid = {'x': slice(None, 5, 5), 'y': slice(-10, None, 5)}

        data = make_regular_2d_ungridded_data()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.array([[1.0, 2.0],  # 3.0],
                              [4.0, 5.0],  # 6.0],
                              [7.0, 8.0],  # 9.0],
                              [10.0, 11.0]])  # 12.0],
        # [13.0, 14.0, 15.0]],

        assert_arrays_almost_equal(cube_out.data, result)

    @istest
    def test_aggregating_single_point_in_one_dimension_lower_bound_edge_case(self):
        """
        Test to document the behaviour of aggregation when a point on a cell's lower bound is taken to be in that cell
          or out of it. If the point is on the lower bound it is taken to be part of that cell.
        """
        grid = {'y': slice(-12.5, 12.5, 5)}

        data = make_dummy_ungridded_data_single_point(lat=-2.50000)

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[0], [0], [1.0], [0], [0]], mask=[[1], [1], [0], [1], [1]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_aggregating_single_point_in_one_dimension_upper_bound_edge_case(self):
        """
        Test to document the behaviour of aggregation when a point on a cell's upper bound is taken to be in that cell
          or out of it. If the point is on the upper bound it is taken to be part of that cell.
        """
        grid = {'y': slice(-12.5, 12.5, 5)}

        data = make_dummy_ungridded_data_single_point(lat=2.50000)

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[0], [0], [0.0], [1.0], [0]], mask=[[1], [1], [1], [0], [1]], fill_value=float('inf'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_aggregating_edge_cases(self):
        """
        Further tests to reinforce the above edge case tests in two dimensions. Note that the upper edges of the
        ungridded data points array are removed from the aggregate because they fall on the respective cells' upper
        bounds.
        """
        grid = {'x': slice(-5, 5, 5), 'y': slice(-10, 10, 5)}

        data = make_regular_2d_ungridded_data()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.array([[1.0, 2.0],  # 3.0],
                              [4.0, 5.0],  # 6.0],
                              [7.0, 8.0],  # 9.0],
                              [10.0, 11.0]])  # 12.0],
        # [13.0, 14.0, 15.0]],

        assert_arrays_almost_equal(cube_out.data, result)

    @istest
    def test_aggregating_simple_dataset_in_two_dimensions_with_missing_values(self):
        grid = {'x': slice(-7.5, 7.5, 5), 'y': slice(-12.5, 12.5, 5)}

        data = make_regular_2d_ungridded_data_with_missing_values()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[1.0, 2.0, 3.0],
                                 [4.0, 5.0, 6.0],
                                 [7.0, 8.0, 9.0],
                                 [10.0, 11.0, 12.0],
                                 [13.0, 14.0, 15.0]],
                                mask=[[0, 0, 0],
                                      [0, 1, 0],
                                      [0, 0, 1],
                                      [0, 0, 0],
                                      [1, 0, 0]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_mean_kernel_with_dataset_in_two_dimensions_with_missing_values(self):
        grid = {'x': slice(-7.5, 7.5, 5), 'y': slice(-12.5, 12.5, 12.5)}

        data = make_regular_2d_ungridded_data_with_missing_values()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[2.5, 2.0, 4.5],
                                 [8.5, 11.0, 13.5]],
                                mask=[[0, 0, 0],
                                      [0, 0, 0]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_max_kernel_with_dataset_in_two_dimensions_with_missing_values(self):
        self.kernel = max()

        grid = {'x': slice(-7.5, 7.5, 5), 'y': slice(-12.5, 12.5, 12.5)}

        data = make_regular_2d_ungridded_data_with_missing_values()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[4.0, 2.0, 6.0],
                                 [10.0, 14.0, 15.0]],
                                mask=[[0, 0, 0],
                                      [0, 0, 0]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_min_kernel_with_dataset_in_two_dimensions_with_missing_values(self):
        self.kernel = min()

        grid = {'x': slice(-7.5, 7.5, 5), 'y': slice(-12.5, 12.5, 12.5)}

        data = make_regular_2d_ungridded_data_with_missing_values()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[1.0, 2.0, 3.0],
                                 [7.0, 8.0, 12.0]],
                                mask=[[0, 0, 0],
                                      [0, 0, 0]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    @istest
    def test_stddev_kernel_with_dataset_in_two_dimensions_with_missing_values(self):
        self.kernel = stddev()

        grid = {'x': slice(-7.5, 7.5, 5), 'y': slice(-12.5, 12.5, 12.5)}

        data = make_regular_2d_ungridded_data_with_missing_values()

        cube_out = data.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[numpy.sqrt(4.5), float('NaN'), numpy.sqrt(4.5)],
                                 [numpy.sqrt(4.5), 3.0, numpy.sqrt(4.5)]],
                                mask=[[0, 1, 0],
                                      [0, 0, 0]], fill_value=float('nan'))

        compare_masked_arrays(cube_out.data, result)

    def test_aggregation_one_dim_using_moments_kernel(self):
        data = make_regular_2d_ungridded_data_with_missing_values()
        grid = {'y': slice(-12.5, 12.5, 12.5)}

        output = data.aggregate(how='moments', **grid)

        expect_mean = numpy.array([[3.2], [11]])
        expect_stddev = numpy.array([[numpy.sqrt(3.7)], [numpy.sqrt(26 / 3.0)]])
        expect_count = numpy.array([[5], [7]])

        assert isinstance(output, GriddedDataList)
        assert len(output) == 3
        actual_mean, actual_stddev, actual_count = output
        assert actual_mean.var_name == 'rain'
        assert actual_stddev.var_name == 'rain_std_dev'
        assert actual_count.var_name == 'rain_num_points'
        assert_arrays_almost_equal(actual_mean.data, expect_mean)
        assert_arrays_almost_equal(actual_stddev.data, expect_stddev)
        assert_arrays_almost_equal(actual_count.data, expect_count)

    def test_aggregation_two_dims_using_moments_kernel(self):
        data = make_regular_2d_ungridded_data_with_missing_values()
        grid = {'y': slice(-12.5, 12.5, 15), 'x': slice(-7.5, 7.5, 10)}

        output = data.aggregate(how='moments', **grid)

        expect_mean = numpy.array([[4.4, 4.5], [35.0 / 3, 13.5]])
        expect_stddev = numpy.array([[numpy.sqrt(9.3), numpy.sqrt(4.5)],
                                     [numpy.sqrt(13.0 / 3), numpy.sqrt(4.5)]])
        expect_count = numpy.array([[5, 2], [3, 2]])

        assert isinstance(output, GriddedDataList)
        assert len(output) == 3
        actual_mean, actual_stddev, actual_count = output
        assert actual_mean.var_name == 'rain'
        assert actual_stddev.var_name == 'rain_std_dev'
        assert actual_count.var_name == 'rain_num_points'
        assert_arrays_almost_equal(actual_mean.data, expect_mean)
        assert_arrays_almost_equal(actual_stddev.data, expect_stddev)
        assert_arrays_almost_equal(actual_count.data, expect_count)

    def test_aggregating_on_grid_0_to_360_when_data_is_minus_180_to_180(self):
        data = make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=9, lon_min=-175., lon_max=145.)
        grid = {'x': slice(125, 270, 40)}
        output = data.aggregate(how=self.kernel, **grid)
        assert_arrays_equal(output.data, [[13.5, 5.5, 6.5, 7.5]])

    def test_aggregating_on_grid_minus_180_to_180_when_data_is_0_to_360(self):
        data = make_regular_2d_ungridded_data(lat_dim_length=2, lon_dim_length=9, lon_min=5., lon_max=325.)
        grid = {'x': slice(-75, 125, 40)}
        output = data.aggregate(how=self.kernel, **grid)
        assert_arrays_equal(output.data, [[12.5, 13.5, 5.5, 6.5, 7.5]])

    def test_collapsed_coords_get_output_as_length_1(self):
        data = make_regular_2d_ungridded_data()
        grid = {'x': slice(0, 360, 10)}
        output = data.aggregate(how=self.kernel, **grid)
        lat = output.coord('latitude')
        assert_that(lat.points, is_([0]))

    def test_collapsed_coords_get_max_min_bounds(self):
        data = make_regular_2d_ungridded_data()
        grid = {'y': slice(-90, 90, 10)}
        output = data.aggregate(how=self.kernel, **grid)
        lon = output.coord('longitude')
        assert_arrays_equal(lon.bounds, [[-5, 5]])

    def test_aggregating_coord_to_length_one_with_explicit_bounds_gets_output_as_length_one(self):
        data = make_regular_2d_ungridded_data()
        grid = {'x': slice(-180, 180, 360), 'y': slice(-90, 90, 10),}
        output = data.aggregate(how=self.kernel, **grid)
        lon = output.coord('longitude')
        assert_that(lon.points, is_([0]))

    def test_aggregating_to_length_one_with_explicit_bounds_get_correct_bounds(self):
        data = make_regular_2d_ungridded_data()
        grid = {'x': slice(-180, 180, 360), 'y': slice(-90, 90, 10),}
        output = data.aggregate(how=self.kernel, **grid)
        lon = output.coord('longitude')
        assert_arrays_equal(lon.bounds, [[-180, 180]])

    def test_aggregating_over_time(self):
        from datetime import datetime, timedelta
        from cis.time_util import cis_standard_time_unit as tunit
        data = make_regular_2d_with_time_ungridded_data()
        data.time.convert_to_std_time()
        output = data.aggregate(t=slice(datetime(1984, 8, 1), datetime(1984, 9, 30), timedelta(days=30)))
        expected_t_bounds = [[tunit.date2num(datetime(1984, 8, 1)), tunit.date2num(datetime(1984, 8, 31))],
                             [tunit.date2num(datetime(1984, 8, 31)), tunit.date2num(datetime(1984, 9, 30))]]
        assert_arrays_equal(output[0].coord('time').bounds, expected_t_bounds)
        assert_arrays_equal(output[0].data, [[[2.5, 10]]])

    def test_aggregating_over_time_with_default_times(self):
        from datetime import datetime, timedelta
        from cis.time_util import cis_standard_time_unit as tunit
        data = make_regular_2d_with_time_ungridded_data()
        data.time.convert_to_std_time()
        output = data.aggregate(t=slice(None, None, timedelta(days=30)))
        expected_t_bounds = [[tunit.date2num(datetime(1984, 8, 27)), tunit.date2num(datetime(1984, 9, 10))]]
        assert_arrays_equal(output[0].coord('time').bounds, expected_t_bounds)
        assert_arrays_equal(output[0].data, [[[7.5]]])

    def test_aggregating_over_time_with_partial_datetime(self):
        from cis.time_util import PartialDateTime, cis_standard_time_unit as tunit
        from datetime import datetime, timedelta
        data = make_regular_2d_with_time_ungridded_data()
        data.time.convert_to_std_time()
        output = data.aggregate(t=[PartialDateTime(1984,9), timedelta(days=30)])
        expected_t_bounds = [[tunit.date2num(datetime(1984, 9, 1)), tunit.date2num(datetime(1984, 10, 1))]]
        assert_arrays_almost_equal(output[0].coord('time').bounds, expected_t_bounds)
        assert_arrays_almost_equal(output[0].data, [[[10.5]]])

    @raises(ValueError)
    def test_empty_step_raises_error_with_partial_datetime(self):
        from cis.time_util import PartialDateTime
        data = make_regular_2d_with_time_ungridded_data()
        data.time.convert_to_std_time()
        output = data.aggregate(t=[PartialDateTime(1984, 9), None])

    @raises(ValueError)
    def test_empty_step_raises_error(self):
        data = make_regular_2d_with_time_ungridded_data()
        data.time.convert_to_std_time()
        output = data.aggregate(x=slice(1, 2, None))


class TestUngriddedListAggregation(TestCase):
    def setUp(self):
        self.kernel = mean()

    @istest
    def test_aggregating_list_of_datasets_over_two_dims(self):
        grid = {'x': slice(-7.5, 7.5, 5), 'y': slice(-12.5, 12.5, 5)}

        datalist = UngriddedDataList([make_regular_2d_ungridded_data_with_missing_values(),
                                      make_regular_2d_ungridded_data_with_missing_values()])

        cube_out = datalist.aggregate(how=self.kernel, **grid)

        result = numpy.ma.array([[1.0, 2.0, 3.0],
                                 [4.0, 5.0, 6.0],
                                 [7.0, 8.0, 9.0],
                                 [10.0, 11.0, 12.0],
                                 [13.0, 14.0, 15.0]],
                                mask=[[0, 0, 0],
                                      [0, 1, 0],
                                      [0, 0, 1],
                                      [0, 0, 0],
                                      [1, 0, 0]], fill_value=float('nan'))

        assert len(cube_out) == 2
        compare_masked_arrays(cube_out[0].data, result)
        compare_masked_arrays(cube_out[1].data, result)

    def test_aggregation_one_dim_using_moments_kernel(self):
        self.kernel = moments()
        data1 = make_regular_2d_ungridded_data_with_missing_values()
        data2 = make_regular_2d_ungridded_data_with_missing_values()
        data2.metadata._name = 'snow'
        data2._data += 10
        data = UngriddedDataList([data1, data2])
        grid = {'y': slice(-12.5, 12.5, 12.5)}

        output = data.aggregate(how=self.kernel, **grid)

        expect_mean = numpy.array([[3.2], [11]])
        expect_stddev = numpy.array([[numpy.sqrt(3.7)], [numpy.sqrt(26.0 / 3)]])
        expect_count = numpy.array([[5], [7]])

        assert isinstance(output, GriddedDataList)
        assert len(output) == 6
        mean_1, stddev_1, count_1, mean_2, stddev_2, count_2 = output
        assert mean_1.var_name == 'rain'
        assert stddev_1.var_name == 'rain_std_dev'
        assert count_1.var_name == 'rain_num_points'
        assert mean_2.var_name == 'snow'
        assert stddev_2.var_name == 'snow_std_dev'
        assert count_2.var_name == 'snow_num_points'
        assert_arrays_almost_equal(mean_1.data, expect_mean)
        assert_arrays_almost_equal(stddev_1.data, expect_stddev)
        assert_arrays_almost_equal(count_1.data, expect_count)
        assert_arrays_almost_equal(mean_2.data, expect_mean + 10)
        assert_arrays_almost_equal(stddev_2.data, expect_stddev)
        assert_arrays_almost_equal(count_2.data, expect_count)

    def test_aggregation_two_dims_using_moments_kernel(self):
        self.kernel = moments()
        data1 = make_regular_2d_ungridded_data_with_missing_values()
        data2 = make_regular_2d_ungridded_data_with_missing_values()
        data2.metadata._name = 'snow'
        data2._data += 10
        data = UngriddedDataList([data1, data2])
        grid = {'y': slice(-12.5, 12.5, 15), 'x': slice(-7.5, 7.5, 10)}

        output = data.aggregate(how=self.kernel, **grid)

        expect_mean = numpy.array([[4.4, 4.5], [35.0 / 3, 13.5]])
        expect_stddev = numpy.array([[numpy.sqrt(9.3), numpy.sqrt(4.5)],
                                     [numpy.sqrt(13.0 / 3), numpy.sqrt(4.5)]])
        expect_count = numpy.array([[5, 2], [3, 2]])

        assert isinstance(output, GriddedDataList)
        assert len(output) == 6
        mean_1, stddev_1, count_1, mean_2, stddev_2, count_2 = output
        assert mean_1.var_name == 'rain'
        assert stddev_1.var_name == 'rain_std_dev'
        assert count_1.var_name == 'rain_num_points'
        assert mean_2.var_name == 'snow'
        assert stddev_2.var_name == 'snow_std_dev'
        assert count_2.var_name == 'snow_num_points'
        assert_arrays_almost_equal(mean_1.data, expect_mean)
        assert_arrays_almost_equal(stddev_1.data, expect_stddev)
        assert_arrays_almost_equal(count_1.data, expect_count)
        assert_arrays_almost_equal(mean_2.data, expect_mean + 10)
        assert_arrays_almost_equal(stddev_2.data, expect_stddev)
        assert_arrays_almost_equal(count_2.data, expect_count)

    @istest
    def test_aggregating_list_of_datasets_over_two_dims_with_diff_masks(self):
        grid = {'x': slice(-7.5, 7.5, 5), 'y': slice(-12.5, 12.5, 5)}

        var_0 = make_regular_2d_ungridded_data_with_missing_values()
        var_1 = make_regular_2d_ungridded_data_with_missing_values()

        var_1.data.mask = 1

        datalist = UngriddedDataList([var_0, var_1])

        cube_out = datalist.aggregate(how=self.kernel, **grid)

        result_0 = numpy.ma.array([[1.0, 2.0, 3.0],
                                 [4.0, 5.0, 6.0],
                                 [7.0, 8.0, 9.0],
                                 [10.0, 11.0, 12.0],
                                 [13.0, 14.0, 15.0]],
                                mask=[[0, 0, 0],
                                      [0, 1, 0],
                                      [0, 0, 1],
                                      [0, 0, 0],
                                      [1, 0, 0]], fill_value=float('nan'))

        result_1 = numpy.ma.array([[1.0, 2.0, 3.0],
                                   [4.0, 5.0, 6.0],
                                   [7.0, 8.0, 9.0],
                                   [10.0, 11.0, 12.0],
                                   [13.0, 14.0, 15.0]],
                                  mask=[[1, 1, 1],
                                        [1, 1, 1],
                                        [1, 1, 1],
                                        [1, 1, 1],
                                        [1, 1, 1]], fill_value=float('nan'))

        print(cube_out[0].data.fill_value)
        assert len(cube_out) == 2
        compare_masked_arrays(cube_out[0].data, result_0)
        compare_masked_arrays(cube_out[1].data, result_1)



import fnmatch
import logging

from cis.data_io.gridded_data import GriddedDataList
from cis.data_io.ungridded_data import UngriddedDataList
from cis.data_io.products.AProduct import get_data, get_coordinates, get_variables
from cis.utils import listify


def expand_filelist(filelist):
    """
    :param filelist: A single element, or list, or comma seperated string of filenames, wildcarded filenames or
     directories
    :return: A flat list of files which exist - with no duplicates
    :raises ValueError: if any of the files in the list do not exist.
    """
    import os
    import six
    from glob import glob
    from cis.utils import OrderedSet

    if isinstance(filelist, six.string_types):
        input_list = filelist.split(',')
    else:
        input_list = listify(filelist)

    # Ensure we don't get duplicates by making file_set a set
    file_set = OrderedSet()
    for element in input_list:
        if any(wildcard in element for wildcard in ['*', '?', ']', '}']):
            filelist = glob(element)
            filelist.sort()
            for filename in filelist:
                file_set.add(filename)
        elif os.path.isdir(element):
            filelist = os.listdir(element)
            filelist.sort()
            for a_file in filelist:
                full_file = os.path.join(element, a_file)
                if os.path.isfile(full_file):
                    file_set.add(full_file)
        elif os.path.isfile(element):
            file_set.add(element)
        else:
            raise ValueError("{} is not a valid filename".format(element))
    return list(file_set)


class DataReader(object):
    """
    High level class to manage reading data from a file.
    Principally, manages operations between one or multiple variables, and gridded or un-gridded data.
    """

    def __init__(self, get_data_func=get_data, get_coords_func=get_coordinates, get_variables_func=get_variables):
        """
        Construct a new DataReader object

        :param get_data_func: Function to read data from file and return a CommonDataList
        :param get_coords_func: Function to read data from a file and return a CoordList
        :param get_variables_func: Function to read variables from a file and return a list of variable strings
        """
        self._get_data_func = get_data_func
        self._get_coords_func = get_coords_func
        self._get_vars_func = get_variables_func

    def read_data_list(self, filenames, variables, product=None, aliases=None):
        """
        Read multiple data objects. Files can be either gridded or ungridded but not a mix of both.

        :param filenames: One or more filenames of the files to read
        :type filenames: string or list
        :param variables: One or more variables to read from the files
        :type variables: string or list
        :param str product: Name of data product to use (optional)
        :param aliases: List of variable aliases to put on each variables
         data object as an alternative means of identifying them. (Optional)
        :return:  A list of the data read out (either a GriddedDataList or UngriddedDataList depending on the
         type of data contained in the files)
        """
        # if filenames or variables are not lists, make them lists of 1 element
        filenames = listify(filenames)
        variables = listify(variables)
        aliases = listify(aliases) if aliases else None

        variables = self._expand_wildcards(variables, filenames, product)

        data_list = None
        for idx, variable in enumerate(variables):
            var_data = self._get_data_func(filenames, variable, product)
            var_data.filenames = filenames
            if aliases:
                try:
                    var_data.alias = aliases[idx]
                except IndexError:
                    raise ValueError("Number of aliases does not match number of variables")
            if data_list is None:
                data_list = GriddedDataList() if var_data.is_gridded else UngriddedDataList()
            data_list.append(var_data)
        assert data_list is not None
        return data_list

    def _expand_wildcards(self, variables, filenames, product):
        """
        Convert any wildcards into actual variable names by inspecting the file

        :param variables: List of variables names, some of which may be wilcards
        :param filenames: Filenames
        :return: List of variable names to use.
        """
        valid_vars = []
        for variable in variables:
            is_wildcard = any(wildcard in variable for wildcard in ['*', '?', ']', '}'])
            file_variables = None
            if is_wildcard:
                if file_variables is None:
                    file_variables = self._get_vars_func(filenames, product)
                matches = fnmatch.filter(file_variables, variable)
                if len(matches) == 0:
                    logging.warning("No variables matching wildcard '%s' found in file." % variable)
                else:
                    logging.info("Wildcard '%s' found matching variables: %s" % (variable, matches))
                valid_vars.extend(matches)
            else:
                valid_vars.append(variable)
        if len(valid_vars) == 0:
            raise ValueError("No matching variables could be found for the variables supplied")
        return valid_vars

    def read_datagroups(self, datagroups):
        """
        Read data from a set of datagroups

        :param datagroups: A list of datagroups. Each datagroup represents a grouping of files and variables, where the
         set of files may be logically considered to represent the same data (an example would be 2D model data split
         into monthly output files where the grid is the same).
         The following should be true of a datagroup:

         1. All variables in a datagroup are present in all the files in that datagroup
         2. The shape of the data returned from each variable must be the same in each file, so that they may be
            concatenated
         3. They should all be openable by the same CIS data product
         4. They should be dictionaries of the following format::

                 {'filenames': ['filename1.nc', 'filename2.nc'],
                   'variables': ['variable1', 'variable2'],
                   'product' : 'Aerosol_CCI_L2'}

        :return list: A list of CommonData objects (either GriddedData or UngriddedData, *or a combination*)
        """
        data_list = list()
        for datagroup in datagroups:
            data_list.extend(self.read_single_datagroup(datagroup))
        return data_list

    def read_single_datagroup(self, datagroup):
        """
        Read data from a set of datagroups

        :param datagroup: A datagroup: a grouping of files and variables, where the
         set of files may be logically considered to represent the same data (an example would be 2D model data split
         into monthly output files where the grid is the same).
         The following should be true of a datagroup:

         1. All variables in a datagroup are present in all the files in that datagroup
         2. The shape of the data returned from each variable must be the same in each file, so that they may be
            concatenated
         3. They should all be openable by the same CIS data product
         4. They should be dictionaries of the following format::

                 {'filenames': ['filename1.nc', 'filename2.nc'],
                   'variables': ['variable1', 'variable2'],
                   'product' : 'Aerosol_CCI_L2'}

        :return CommonDataList: Either a GriddedDataLise or an UngriddedDataList
        """
        aliases = datagroup.get('aliases', None)
        data = self.read_data_list(datagroup['filenames'], datagroup['variables'],
                                   datagroup.get('product', None), aliases)
        return data

    def read_coordinates(self, filenames, product=None):
        """
        Read the coordinates from a file
        :param filenames:   The filename of the files to read
        :return: A CoordList object
        """

        # if filenames is not a list, make it a list of 1 element
        filenames = listify(filenames)

        return self._get_coords_func(filenames, product)



from hamcrest import assert_that, is_
import numpy


def assert_arrays_equal(result, expected):
    assert_that(numpy.array_equal(result, expected), is_(True),
                "arrays not the same. Expected\n {}\n was\n {}".format(expected, result))


def assert_arrays_almost_equal(result, expected, tol=1.0e-15):
    assert_that(numpy.allclose(result, expected, atol=tol)
                , is_(True),
                "arrays not almost the same. Expected\n {}\n was\n {}".format(expected, result))


def compare_masked_arrays(a1, a2):
    """
    Compare two masked arrays:
    - Masks should be the same
    - Unmasked data should be same
    - Shape should be same
    - Numeric values that are 'masked out' don't matter
    """
    flat_1 = a1.compressed()
    flat_2 = a2.compressed()
    assert_that(numpy.allclose(flat_1, flat_2), 'Masked arrays have different values')
    assert_that(numpy.array_equal(a1.mask, a2.mask), 'Masked arrays have different masks')
    assert_that(a1.shape, is_(a2.shape), 'Masked arrays have different shapes')



"""
Routines for creating a plot and then formatting it, using command line options. It is not intended for plotting
directly from Python, although it could be used for that.
"""


def set_log_scales(ax, logx, logy, rescale=True):
    """
    Optionally log-scale one or both of the axis
    """
    if logx:
        ax.set_xscale("log")
    if logy:
        ax.set_yscale("log")

    if (logx or logy) and rescale:
        # Optionally rescale the data as it doesn't appear to be done automatically.
        ax.relim()
        ax.autoscale()


def format_plot(ax, grid, fontsize, xlabel, ylabel, title):
    """
    General high-level level formatting
    """
    import matplotlib

    if grid:
        ax.grid(True, which="both")

    if fontsize is not None:
        matplotlib.rcParams.update({'font.size': fontsize})

    if xlabel is not None:
        ax.set_xlabel(xlabel)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if title is not None:
        ax.set_title(title)


def apply_map_axis_limits(ax, xmin=None, xmax=None, xstep=None, ymin=None, ymax=None, ystep=None):
    """
    Applies the specified limits to the given GeoAxis.
    """
    import cartopy.crs as ccrs
    from cis.plotting.plot import get_best_map_ticks
    import numpy as np

    transform = ccrs.PlateCarree(360)

    global_tolerance = 0.8

    # We can't optionally pass in certain bounds to set_extent so we need to pull out the existing ones and only
    #  change the ones we've been given.
    x1, x2, y1, y2 = ax.get_extent()
    # If the user hasn't specified any limits and the data spans most of the globe, just make it a global plot
    if all(lim is None for lim in (xmin, xmax, ymin, ymax)) and \
            ((y2 - y1 > (ax.projection.y_limits[1] - ax.projection.y_limits[0]) * global_tolerance) or
                 (x2 - x1 > (ax.projection.x_limits[1] - ax.projection.x_limits[0]) * global_tolerance)):
        ax.set_global()
    else:
        xmin = xmin if xmin is not None else x1
        xmax = xmax if xmax is not None else x2
        ymin = ymin if ymin is not None else y1
        ymax = ymax if ymax is not None else y2
        ax.set_extent([xmin, xmax, ymin, ymax], crs=transform)

    # Get the updated extent
    x1, x2, y1, y2 = ax.get_extent()

    # Get default ticks
    xticks, yticks = get_best_map_ticks(ax)

    # If we're given user steps then calculate our own ticks
    if xstep is not None:
        xticks = np.arange(x1, x2 + xstep, xstep)

    if ystep is not None:
        yticks = np.arange(y1, y2 + ystep, ystep)

    ax.set_xticks(xticks)
    ax.set_yticks(yticks)


def apply_axis_limits(ax, xmin=None, xmax=None, xstep=None, ymin=None, ymax=None, ystep=None):
    """
    Applies the specified limits to the given axis
    """
    import numpy as np

    ax.set_xlim(xmin=xmin, xmax=xmax)
    ax.set_ylim(ymin=ymin, ymax=ymax)

    if xstep is not None:
        min_val, max_val = ax.get_xlim()
        ticks = np.arange(min_val, max_val + xstep, xstep)

        ax.set_xticks(ticks)

    if ystep is not None:
        min_val, max_val = ax.get_ylim()
        ticks = np.arange(min_val, max_val + ystep, ystep)

        ax.set_yticks(ticks)


def get_x_wrap_start(data_list, user_xmin=None):
    """
    Find the left-hand most data point out of a list of CommonData objects. This is needed to work out the best
     central longitude.

    If the user has specified an xmin then if it is -ve take our wrap start as -180, else it should be 0.
    :param list data_list: CommonData objects
    :param float user_xmin: An optional user supplied x_min.
    :return: The left-hand most longitude of the plot.
    """
    from cis.utils import find_longitude_wrap_start as find_start
    from datetime import datetime

    # FIND THE WRAP START OF THE DATA
    all_starts = [find_start(data) for data in data_list if find_start(data) is not None]
    data_wrap_start = min(all_starts) if all_starts else None

    # NOW find the wrap start of the user specified range
    if user_xmin is not None and not isinstance(user_xmin, datetime):
        x_wrap_start = -180 if user_xmin < 0 else 0
    else:
        x_wrap_start = data_wrap_start

    return x_wrap_start


class Plotter(object):

    def __init__(self, data, type=None, output=None, height=None,
                 width=None, logx=False, logy=False, xmin=None,
                 xmax=None, xstep=None, ymin=None, ymax=None, ystep=None, nasabluemarble=False,
                 grid=False, xlabel=None, ylabel=None, title=None, fontsize=None, *args, **kwargs):
        """
        Constructor for the formatted plotter. Note that this method also does the actual plotting.

        :param list data: A list of packed (i.e. GriddedData or UngriddedData objects) data items to be plotted
        :param string type: The plot type to be used
        :param string output: The filename of the file to save the plot to (Optional). Various file extensions can be
         used, with png being the default.
        :param float height: The height of the plot
        :param float width: The width of the plot
        :param bool logx: Set the x-axis to log scale
        :param bool logy: Set the y-axis to log scale
        :param float xmin: The minimium value of the x axis
        :param float xmax: The maximium value of the x axis
        :param float xstep: The step value of the x ticks
        :param float ymin: The minimium value of the y axis
        :param float ymax: The maximium value of the y axis
        :param float ystep: The step value of the y ticks
        :param bool nasabluemarble: Plot a NASA bluemarble background? Default False
        :param bool grid: Plot gridlines? Default False
        :param string xlabel: A label for the x axis
        :param string ylabel: A label for the y axis
        :param string title: A title for the plot
        :param int fontsize: The fontsize to use for the ticks and labels
        :param args: Any other arguments received from the parser
        :param kwargs: Any other keyword arguments received from the parser
        """
        from cis.plotting.plot import multilayer_plot, basic_plot, drawbluemarble

        # Figure out teh best coneral longitude for the projection (it doesn't matter if it's a map yet)
        x_start = get_x_wrap_start(data, xmin)
        if x_start is not None and 'central_longitude' not in kwargs:
            kwargs['central_longitude'] = x_start - 180.0

        # Turn data into a single object if it is one - otherwise we end up with an overlay plot
        if isinstance(data, list) and len(data) == 1:
            data = data[0]

        # If it's still a list... We don't use the object methods because in the case of the command line API
        #  we allow mixed Gridded and Ungridded data sets - which we don't allow for CommonDataLists
        if isinstance(data, list):
            plot, self.ax = multilayer_plot(data, how=type, *args, **kwargs)
        else:
            if 'layer_opts' in kwargs:
                kwargs.update(kwargs.pop('layer_opts')[0])
            plot, self.ax = basic_plot(data, how=type, *args, **kwargs)

        self.fig = self.ax.get_figure()

        self.set_width_and_height(width, height)

        plot.set_log_scales(self.ax, logx, logy)

        format_plot(self.ax, grid, fontsize, xlabel, ylabel, title)

        if plot.is_map():
            if plot.is_rectangular_projection(self.ax):
                apply_map_axis_limits(self.ax, xmin, xmax, xstep, ymin, ymax, ystep)
            # This has to come after applying the axis limits because otherwise the image can get cropped
            if nasabluemarble:
                drawbluemarble(self.ax)
        elif type != 'taylor':
            apply_axis_limits(self.ax, xmin, xmax, xstep, ymin, ymax, ystep)

        self.output_to_file_or_screen(output)

    def output_to_file_or_screen(self, out_filename=None):
        """
        Outputs to screen unless a filename is given
        """
        import logging
        import matplotlib.pyplot as plt

        if out_filename is None:
            plt.show()
        else:
            logging.info("saving plot to file: " + out_filename)
            width = self.fig.get_figwidth()
            self.fig.savefig(out_filename, bbox_inches='tight',
                             pad_inches=0.05 * width)  # Will overwrite if file already exists

    def set_width_and_height(self, width, height):
        """
        Sets the width and height of the plot
        Uses an aspect ratio of 4:3 if only one of width and height are specified
        If neither width or height are specified it defaults to 8 by 6 inches.
        """

        if height is not None:
            if width is None:
                width = height * (4.0 / 3.0)
        elif width is not None:
            height = width * (3.0 / 4.0)
        else:
            height = 6
            width = 8

        self.fig.set_figheight(height)
        self.fig.set_figwidth(width)


from time import gmtime, strftime
import logging

import iris
import numpy as np

from cis.data_io.common_data import CommonData, CommonDataList
from cis.data_io.hyperpoint import HyperPoint
from cis.data_io.hyperpoint_view import GriddedHyperPointView
import six


def load_cube(*args, **kwargs):
    """
    Load a single GriddedData object through the iris load interface, but also attempt concatenation if merging fails

    :return GriddedData: A single GriddedData object
    :raises ValueError: If 0 or more than one cube is found
    """
    from iris.exceptions import MergeError, ConcatenateError

    cubes = iris.load(*args, **kwargs)

    try:
        iris_cube = cubes.merge_cube()
    except MergeError as e:
        logging.info("Unable to merge cubes on load: \n {}\nAttempting to concatenate instead.".format(e))
        try:
            iris_cube = cubes.concatenate_cube()
        except ConcatenateError as e:
            logging.error("Unable to concatenate cubes on load: \n {}".format(e))
            raise ValueError("Unable to create a single cube from arguments given: {}".format(args))
    except ValueError as e:
        raise ValueError("No cubes found")
    return make_from_cube(iris_cube)


def make_from_cube(cube):
    gd = None
    if isinstance(cube, iris.cube.Cube):
        return GriddedData.make_from_cube(cube)
    elif isinstance(cube, iris.cube.CubeList):
        return GriddedDataList(cube)
    return gd


class GriddedData(iris.cube.Cube, CommonData):

    def __init__(self, *args, **kwargs):

        try:
            standard_name = kwargs['standard_name']
            try:
                iris.std_names.STD_NAMES[standard_name]
            except KeyError:
                rejected_name = kwargs.pop('standard_name')
                logging.warning("Standard name '{}' not CF-compliant, this standard name will not be "
                                "used in the output file.".format(rejected_name))
        except KeyError:
            pass

        self._local_attributes = []

        try:
            super(GriddedData, self).__init__(*args, **kwargs)
        except ValueError:
            rejected_unit = kwargs.pop('units')
            logging.warning("Attempted to set invalid unit '{}'.".format(rejected_unit))
            super(GriddedData, self).__init__(*args, **kwargs)

    @classmethod
    def make_from_cube(cls, cube):
        """
        Create a GriddedData object from a cube
        :param cube:
        :return:
        """
        if not isinstance(cube, GriddedData):
            cube.__class__ = GriddedData
            cube._local_attributes = []
        return cube

    def make_new_with_same_coordinates(self, data=None, var_name=None, standard_name=None,
                                       long_name=None, history=None, units=None, flatten=False):
        """
        Create a new, empty GriddedData object with the same coordinates as this one
        :param data: Data to use (if None then defaults to all zeros)
        :param var_name: Variable name
        :param standard_name: Variable CF standard name
        :param long_name: Variable long name
        :param history: Data history string
        :param units: Variable units
        :param flatten: Whether to flatten the data shape (for ungridded data only)
        :return: GriddedData instance
        """
        if data is None:
            data = np.zeros(self.shape)
        data = GriddedData(data=data, standard_name=standard_name, long_name=long_name, var_name=var_name,
                           units=units, dim_coords_and_dims=self._dim_coords_and_dims,
                           aux_coords_and_dims=self._aux_coords_and_dims, aux_factories=self._aux_factories)
        # Add history separately as it is not a constructor argument
        data.add_history(history)
        return data

    def __repr__(self):
        # Override the Cube representation
        return "<cis 'GriddedData' of %s>" % self.summary(shorten=True)

    @staticmethod
    def _wrap_cube_iterator(itr):
        """Makes a generator that returns a GriddedData object from each Cube returned by an iterator.
        :param itr: iterator over Cubes
        :return: yields GriddedData objects created from Cubes
        """
        for c in itr:
            yield make_from_cube(c)

    def slices(self, *args, **kwargs):
        return self._wrap_cube_iterator(super(GriddedData, self).slices(*args, **kwargs))

    def get_coordinates_points(self):
        """Returns a HyperPointView of the points.
        :return: HyperPointView of all the data points
        """
        all_coords = [((c[0].points, c[1]) if c is not None else None) for c in self.find_standard_coords()]
        return GriddedHyperPointView(all_coords, self.data)

    def get_all_points(self):
        """Returns a HyperPointView of the points.
        :return: HyperPointView of all the data points
        """
        all_coords = [((c[0].points, c[1]) if c is not None else None) for c in self.find_standard_coords()]
        return GriddedHyperPointView(all_coords, self.data)

    def get_non_masked_points(self):
        """Returns a HyperPointView of the points.
        :return: HyperPointView of all the data points
        """
        all_coords = [((c[0].points, c[1]) if c is not None else None) for c in self.find_standard_coords()]
        return GriddedHyperPointView(all_coords, self.data, non_masked_iteration=True)

    def find_standard_coords(self):
        """Constructs a list of the standard coordinates.
        The standard coordinates are latitude, longitude, altitude, air_pressure and time; they occur in the return
        list in this order.
        :return: list of coordinates or None if coordinate not present
        """
        ret_list = []

        coords = self.coords(dim_coords=True)
        for name in HyperPoint.standard_names:
            coord_and_dim = None
            for idx, coord in enumerate(coords):
                if coord.standard_name == name:
                    coord_and_dim = (coord, idx)
                    break
            ret_list.append(coord_and_dim)

        return ret_list

    @property
    def history(self):
        """
        Return the history attribute
        :return:
        """
        return self.attributes['history']

    def add_history(self, new_history):
        """Appends to, or creates, the history attribute using the supplied history string.

        The new entry is prefixed with a timestamp.
        :param new_history: history string
        """
        timestamp = strftime("%Y-%m-%dT%H:%M:%SZ ", gmtime())
        if 'history' not in self.attributes:
            self.attributes['history'] = timestamp + new_history
        else:
            self.attributes['history'] += '\n' + timestamp + new_history

    @property
    def is_gridded(self):
        """Returns value indicating whether the data/coordinates are gridded.
        """
        return True

    def set_longitude_range(self, range_start):
        """Rotates the longitude coordinate array and changes its values by
        360 as necessary to force the values to be within a 360 range starting
        at the specified value, i.e.,
        range_start <= longitude < range_start + 360

        The data array is rotated correspondingly around the dimension
        corresponding to the longitude coordinate.

        :param range_start: starting value of required longitude range
        """
        lon_coord = self.coords(standard_name="longitude")
        if len(lon_coord) == 0:
            return
        lon_coord = lon_coord[0]
        lon_idx = self.dim_coords.index(lon_coord)
        # Check if there are bounds which we will need to wrap as well
        roll_bounds = (lon_coord.bounds is not None) and (lon_coord.bounds.size != 0)
        idx1 = np.searchsorted(lon_coord.points, range_start)
        idx2 = np.searchsorted(lon_coord.points, range_start + 360.)
        shift = 0
        new_lon_points = None
        new_lon_bounds = None
        if 0 < idx1 < len(lon_coord.points):
            shift = -idx1
            lon_min = lon_coord.points[idx1]
            new_lon_points = np.roll(lon_coord.points, shift, 0)
            # Calculate which indices need 360 adding to them...
            indices_to_shift_value_of = new_lon_points < lon_min
            # ... then, add 360 to all those longitude values
            new_lon_points[indices_to_shift_value_of] += 360.0
            if roll_bounds:
                # If the coordinate has bounds then roll those as well
                new_lon_bounds = np.roll(lon_coord.bounds, shift, 0)
                # And shift all of the bounds (upper and lower) for those points which we had to shift. We can't do the
                # check independently because there may be cases where an upper or lower bound falls outside of the
                # 360 range, we leave those as they are to preserve monotonicity. See e.g.
                # test_set_longitude_bounds_wrap_at_360
                new_lon_bounds[indices_to_shift_value_of] += 360.0
        elif 0 < idx2 < len(lon_coord.points):
            shift = len(lon_coord.points) - idx2
            lon_max = lon_coord.points[idx2]
            new_lon_points = np.roll(lon_coord.points, shift, 0)
            indices_to_shift_value_of = new_lon_points >= lon_max
            new_lon_points[indices_to_shift_value_of] -= 360.0
            if roll_bounds:
                new_lon_bounds = np.roll(lon_coord.bounds, shift, 0)
                # See comment above re preserving monotinicity.
                new_lon_bounds[indices_to_shift_value_of] -= 360.0
        if shift != 0:
            # Ensure we also roll any auxilliary coordinates
            for aux_coord in self.aux_coords:
                # Find all of the data dimensions which the auxiliary coordinate spans...
                dims = self.coord_dims(aux_coord)
                # .. and check if longitude is one of those dimensions
                if lon_idx in dims:
                    # Now roll the axis of the auxiliary coordinate which is associated with the longitude data
                    # dimension: dims.index(lon_idx)
                    new_points = np.roll(aux_coord.points, shift, dims.index(lon_idx))
                    aux_coord.points = new_points
            # Now roll the data itself
            new_data = np.roll(self.data, shift, lon_idx)
            self.data = new_data
            # Put the new coordinates back in their relevant places
            self.dim_coords[lon_idx].points = new_lon_points
            if roll_bounds:
                self.dim_coords[lon_idx].bounds = new_lon_bounds

    def add_attributes(self, attributes):
        """
        Add a variable attribute to this data
        :param attributes: Dictionary of attribute names (keys) and values.
        :return:
        """
        for key, value in list(attributes.items()):
            try:
                self.attributes[key] = value
            except ValueError:
                try:
                    setattr(self, key, value)
                except ValueError as e:
                    logging.warning("Could not set NetCDF attribute '%s' because %s" % (key, e.args[0]))
        # Record that this is a local (variable) attribute, not a global attribute
        self._local_attributes.extend(list(attributes.keys()))

    def remove_attribute(self, key):
        """
        Remove a variable attribute to this data
        :param key: Attribute key to remove
        :return:
        """
        self.attributes.pop(key, None)
        try:
            self._local_attributes.remove(key)
        except ValueError:
            pass

    def save_data(self, output_file):
        """
        Save this data object to a given output file
        :param output_file: Output file to save to.
        """
        logging.info('Saving data to %s' % output_file)
        save_args = {'local_keys': self._local_attributes}
        # If we have a time coordinate then use that as the unlimited dimension, otherwise don't have any
        if self.coords('time'):
            save_args['unlimited_dimensions'] = ['time']
        iris.save(self, output_file, **save_args)

    def as_data_frame(self, copy=True):
        """
        Convert a GriddedData object to a Pandas DataFrame.

        :param copy: Create a copy of the data for the new DataFrame? Default is True.
        :return: A Pandas DataFrame representing the data and coordinates. Note that this won't include any metadata.
        """
        from iris.pandas import as_data_frame
        return as_data_frame(self, copy=copy)

    def collapsed(self, coords, how=None, **kwargs):
        """
        Collapse the dataset over one or more coordinates using CIS aggregation (NOT Iris). This allows multidimensional
         coordinates to be aggregated over as well.
        :param list of iris.coords.Coord or str coords: The coords to collapse
        :param str or iris.analysis.Aggregator how: The kernel to use in the aggregation
        :param kwargs: NOT USED - this is only to match the iris interface.
        :return:
        """
        return _collapse_gridded(self, coords, how)

    def subset(self, **kwargs):
        """
        Subset the data based on the specified constraints. Note that the limits are inclusive.

        The subset region is defined by passing keyword arguments for each dimension to be subset over, each argument
        must be a slice, or have two entries (a maximum and a minimum). Datetime objects can be used to specify upper
        and lower datetime limits, or a single PartialDateTime object can be used to specify a datetime range.

        The keyword keys are used to find the relevant coordinate, they are looked for in order of name, standard_name,
        axis and var_name.

        For example:
            data.subset(x=[0, 80], y=slice(10, 50))

        or:
            data.aggregate(t=PartialDateTime(2008,9))


        A shape keyword can also be supplied as a WKT string or shapely object to subset in lat/lon by an arbitrary
        shape. In this case the lat/lon bounds are taken as the bounding box of the shape. Note that the shape of the
        output remains the same, but a mask is applied to the data over the relevant coordinates. Only 2-dimensional
        datasets are supported at this time.

        :param kwargs: The constraints for each coordinate dimension
        :return CommonData:
        """
        from cis.subsetting.subset import subset, GriddedSubsetConstraint
        return subset(self, GriddedSubsetConstraint, **kwargs)

    def sampled_from(self, data, how='', kernel=None, missing_data_for_missing_sample=True, fill_value=None,
                     var_name='', var_long_name='', var_units='', **kwargs):
        """
        Collocate the CommonData object with another CommonData object using the specified collocator and kernel

        :param CommonData or CommonDataList data: The data to resample
        :param str how: Collocation method (e.g. lin, nn, bin or box)
        :param str or cis.collocation.col_framework.Kernel kernel:
        :param bool missing_data_for_missing_sample: Should missing values in sample data be ignored for collocation?
        :param float fill_value: Value to use for missing data
        :param str var_name: The output variable name
        :param str var_long_name: The output variable's long name
        :param str var_units: The output variable's units
        :return CommonData: The collocated dataset
        """
        from cis.collocation import col_implementations as ci
        from cis.data_io.ungridded_data import UngriddedData, UngriddedDataList
        from cis.collocation.col import collocate, get_kernel

        if isinstance(data, UngriddedData) or isinstance(data, UngriddedDataList):
            col_cls = ci.GeneralGriddedCollocator
            # Bin is the default for ungridded -> gridded collocation
            if how == '' or how == 'bin':
                con = ci.BinnedCubeCellOnlyConstraint()
            elif how == 'box':
                con = ci.SepConstraintKdtree(**kwargs)
            else:
                raise ValueError("Invalid method specified for ungridded -> gridded collocation: " + how)

            # We can have any kernel, default to moments
            kernel = get_kernel(kernel)
        elif isinstance(data, GriddedData) or isinstance(data, GriddedDataList):
            col_cls = ci.GriddedCollocator
            con = None
            if kernel is not None:
                raise ValueError("Cannot specify kernel when method is 'lin' or 'nn'")

            # Lin is the default for gridded -> gridded
            if how == '' or how == 'lin':
                kernel = ci.gridded_gridded_li()
            elif how == 'nn':
                kernel = ci.gridded_gridded_nn()
            else:
                raise ValueError("Invalid method specified for gridded -> gridded collocation: " + how)
        else:
            raise ValueError("Invalid argument, data must be either GriddedData or UngriddedData")

        col = col_cls(missing_data_for_missing_sample=missing_data_for_missing_sample, fill_value=fill_value,
                      var_name=var_name, var_long_name=var_long_name, var_units=var_units)

        return collocate(data, self, col, con, kernel)

    def _get_default_plot_type(self, lat_lon=False):
        if self.ndim == 1:
            return 'line'
        elif self.ndim ==2:
            return 'heatmap'
        else:
            raise ValueError("Unable to determine plot type for data with {} dimensions".format(self.ndim))


class GriddedDataList(iris.cube.CubeList, CommonDataList):
    """
    This class extends iris.cube.CubeList to add functionality needed for CIS to process multiple gridded data.

    It is expected that it will contain cis.data_io.gridded_data.GriddedData instances (which extend
    iris.cube.Cube) rather than IRIS cubes themselves, as there is additional functionality required which is
    present in the GriddedData wrapper.
    """

    def __str__(self):
        return "GriddedDataList: \n%s" % super(GriddedDataList, self).__str__()

    @property
    def is_gridded(self):
        """Returns value indicating whether the data/coordinates are gridded.
        """
        return True

    def append(self, p_object):
        if isinstance(p_object, iris.cube.Cube):
            p_object = make_from_cube(p_object)
        super(GriddedDataList, self).append(p_object)

    def save_data(self, output_file):
        """
        Save data to a given output file
        :param output_file: File to save to
        """
        logging.info('Saving data to %s' % output_file)
        save_args = {}

        # If we have a time coordinate then use that as the unlimited dimension, otherwise don't have any
        if self.coords('time'):
            save_args['unlimited_dimensions'] = ['time']

        iris.save(self, output_file, **save_args)

    def coord(self, *args, **kwargs):
        """
        Call iris.cube.Cube.coord(*args, **kwargs) for the first item of data (assumes all data in list has
        same coordinates)
        :param args:
        :param kwargs:
        :return:
        """
        return self[0].coord(*args, **kwargs)

    def add_aux_coord(self, *args, **kwargs):
        """
        Call iris.cube.Cube.add_aux_coord(*args, **kwargs) for the all items in the data list
        :param args:
        :param kwargs:
        :return:
        """
        for var in self:
            var.add_aux_coord(*args, **kwargs)

    def coord_dims(self, *args, **kwargs):
        """
        Call iris.cube.Cube.coord_dims(*args, **kwargs) for the first item of data (assumes all data in list has
        same coordinates)
        :param args:
        :param kwargs:
        :return:
        """
        return self[0].coord_dims(*args, **kwargs)

    def remove_coord(self, *args, **kwargs):
        """
        Call iris.cube.Cube.remove_coord(*args, **kwargs) for the all items in the data list
        :param args:
        :param kwargs:
        :return:
        """
        for var in self:
            var.remove_coord(*args, **kwargs)

    def add_dim_coord(self, *args, **kwargs):
        """
        Call iris.cube.Cube.add_dim_coord(*args, **kwargs) for the all items in the data list
        :param args:
        :param kwargs:
        :return:
        """
        for var in self:
            var.add_dim_coord(*args, **kwargs)

    def aggregated_by(self, *args, **kwargs):
        """
        Build an aggregated GriddedDataList by calling iris.cube.Cube.aggregated_by(*args, **kwargs)
        for the all items in the data list
        :param args:
        :param kwargs:
        :return: GriddedDataList
        """
        data_list = GriddedDataList()
        for data in self:
            data_list.append(data.aggregated_by(*args, **kwargs))
        return data_list

    def collapsed(self, *args, **kwargs):
        """
        Collapse the dataset over one or more coordinates using CIS aggregation (NOT Iris). This allows multidimensional
         coordinates to be aggregated over as well.
        :param list of iris.coords.Coord or str coords: The coords to collapse
        :param str or iris.analysis.Aggregator how: The kernel to use in the aggregation
        :param kwargs: NOT USED - this is only to match the iris interface.
        :return:
        """
        output = GriddedDataList()
        for data in self:
            output.extend(data.collapsed(*args, **kwargs))
        return output

    def interpolate(self, *args, **kwargs):
        """
        Perform an interpolation over the GriddedDataList using the iris.cube.Cube.interpolate() method
        :param args: Arguments for the Iris interpolate method
        :param kwargs: Keyword arguments for the Iris interpolate method
        :return: Interpolated GriddedDataList
        """
        output = GriddedDataList()
        for data in self:
            output.append(data.interpolate(*args, **kwargs))
        return output

    def regrid(self, *args, **kwargs):
        """
        Perform a regrid over the GriddedDataList using the iris.cube.Cube.regrid() method
        :param args: Arguments for the Iris regrid method
        :param kwargs: Keyword arguments for the Iris regrid method
        :return: Regridded GriddedDataList
        """
        output = GriddedDataList()
        for data in self:
            output.append(data.regrid(*args, **kwargs))
        return output

    def intersection(self, *args, **kwargs):
        """
        Call the iris.cube.Cube.intersection() method over all the cubes in a GriddedDataList
        :param args: Arguments for the Iris intersection method
        :param kwargs: Keyword arguments for the Iris intersection method
        :return: Intersected GriddedDataList or None if no data in intersection
        """
        output = GriddedDataList()
        for data in self:
            new_data = data.intersection(*args, **kwargs)
            if new_data is None:
                return None
            output.append(data.intersection(*args, **kwargs))
        return output

    def extract(self, *args, **kwargs):
        """
        Call the iris.cube.Cube.extract() method over all the cubes in a GriddedDataList
        :param args: Arguments for the Iris extract method
        :param kwargs: Keyword arguments for the Iris extract method
        :return: Extracted GriddedDataList oR None if all data excluded
        """
        output = GriddedDataList()
        for data in self:
            new_data = data.extract(*args, **kwargs)
            if new_data is None:
                return None
            output.append(data.extract(*args, **kwargs))
        return output

    def transpose(self, *args, **kwargs):
        """
        Call the iris.cube.Cube.transpose() method over all the cubes in a GriddedDataList
        :param args: Arguments for the Iris transpose method
        :param kwargs: Keyword arguments for the Iris transpose method
        """
        for data in self:
            data.transpose(*args, **kwargs)

    @property
    def dim_coords(self):
        """
        The dimension coordinates of this data
        """
        # Use the dimensions of the first item since all items should have the same dimensions
        return self[0].dim_coords

    @property
    def aux_coords(self):
        """
        The auxiliary coordinates of this data
        """
        # Use the dimensions of the first item since all items should have the same dimensions
        return self[0].aux_coords

    @property
    def ndim(self):
        """
        The number of dimensions in the data of this list.
        """
        # Use the dimensions of the first item since all items should be the same shape
        return self[0].ndim

    def subset(self, **kwargs):
        from cis.subsetting.subset import subset, GriddedSubsetConstraint
        return subset(self, GriddedSubsetConstraint, **kwargs)


def _collapse_gridded(data, coords, kernel):
    """
    Collapse a GriddedData or GriddedDataList based on the specified grids (currently only collapsing is available)
    :param GriddedData or GriddedDataList data: The data object to aggregate
    :param list of iris.coords.Coord or str coords: The coords to collapse
    :param str or iris.analysis.Aggregator kernel: The kernel to use in the aggregation
    :return:
    """
    from cis.aggregation.collapse_kernels import aggregation_kernels, MultiKernel
    from iris.analysis import Aggregator as IrisAggregator
    from cis.aggregation.gridded_collapsor import GriddedCollapsor
    from cis import __version__
    from cis.utils import listify

    # Ensure the coords are all Coord instances
    coords = [data._get_coord(c) for c in listify(coords)]

    # The kernel can be a string or object, so catch both defaults
    if kernel is None or kernel == '':
        kernel = 'moments'

    if isinstance(kernel, six.string_types):
        kernel_inst = aggregation_kernels[kernel]
    elif isinstance(kernel, (IrisAggregator, MultiKernel)):
        kernel_inst = kernel
    else:
        raise ValueError("Invalid kernel specified: " + str(kernel))

    aggregator = GriddedCollapsor(data, coords)
    data = aggregator(kernel_inst)

    history = "Collapsed using CIS version " + __version__ + \
              "\n variables: " + str(getattr(data, "var_name", "Unknown")) + \
              "\n from files: " + str(getattr(data, "filenames", "Unknown")) + \
              "\n over coordinates: " + ", ".join(c.name() for c in coords) + \
              "\n with kernel: " + str(kernel_inst) + "."
    data.add_history(history)

    return data



import unittest

from cis.parse import parse_args
from cis.cis_main import stats_cmd, col_cmd
from cis.test.integration_test_data import *
from cis.test.integration.base_integration_test import BaseIntegrationTest


class TestStats(BaseIntegrationTest):

    output_vars = [
        "num_points",
        "dataset_mean_1",
        "dataset_mean_2",
        "dataset_stddev_1",
        "dataset_stddev_2",
        "abs_mean",
        "abs_stddev",
        "rel_mean",
        "rel_stddev",
        "spearman",
        "regression_gradient",
        "regression_intercept",
        "regression_r",
        "regression_stderr"]

    def test_Aeronet_wavelength_stats(self):
        # Takes 3s
        args = ['stats', '%s,%s:%s' % ('AOT_500', 'AOT_440', escape_colons(another_valid_aeronet_filename)),
                '-o', self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        stats_cmd(arguments)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, self.output_vars)

    def test_no_output_file(self):
        # Takes 3s
        args = ['stats', '%s,%s:%s' % ('AOT_500', 'AOT_440', escape_colons(another_valid_aeronet_filename))]
        arguments = parse_args(args)
        stats_cmd(arguments)

    def test_ECHAMHAM_wavelength_stats(self):
        # Takes 0.7s
        args = ['stats', "%s,%s:%s" % (valid_echamham_variable_1, valid_echamham_variable_2, escape_colons(valid_echamham_filename)),
                '-o', self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        stats_cmd(arguments)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, self.output_vars)

    def test_collocated_NetCDF_Gridded_onto_GASSP(self):
        # Takes 2s
        # First do a collocation of ECHAMHAM onto GASSP
        sample_file = valid_GASSP_aeroplane_filename
        sample_var = valid_GASSP_aeroplane_variable
        collocator_and_opts = 'nn,variable=%s' % sample_var
        arguments = ['col', '%s:%s' % (valid_echamham_variable_1, escape_colons(valid_echamham_filename)),
                     escape_colons(sample_file) + ':collocator=' + collocator_and_opts,
                     '-o', 'collocated_gassp']
        main_arguments = parse_args(arguments)
        col_cmd(main_arguments)

        # Then do a statistics calculation using the collocated data:
        args = ['stats', "%s:%s" % (valid_echamham_variable_1, 'collocated_gassp.nc'),
                "%s:%s" % (valid_GASSP_aeroplane_variable, escape_colons(valid_GASSP_aeroplane_filename)),
                '-o', self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        stats_cmd(arguments)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, self.output_vars)
        os.remove('collocated_gassp.nc')

    @skip_pyhdf
    def test_CloudSat(self):
        # Takes 140s
        args = ['stats', "%s,%s:%s" % ("RVOD_liq_water_content", "RVOD_ice_water_content",
                                       escape_colons(valid_cloudsat_RVOD_file)),
                '-o', self.OUTPUT_FILENAME]
        arguments = parse_args(args)
        stats_cmd(arguments)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, self.output_vars)


if __name__ == '__main__':
    unittest.main()



import logging
import iris


class GriddedCollapsor(object):

    def __init__(self, data, coords):
        """
        Set up the collapse of a GriddedData set over the given Coords
        :param GriddedData data:
        :param list coords: of Coord instances
        """
        self.data = data
        self.coords = coords

    @staticmethod
    def _partially_collapse_multidimensional_coord(coord, dims_to_collapse, kernel=iris.analysis.MEAN):
        import operator
        from functools import reduce
        import numpy as np

        # First calculate our new shape and dims
        dims_to_collapse = sorted(dims_to_collapse)
        end_size = reduce(operator.mul, (coord.shape[dim] for dim in dims_to_collapse))
        untouched_dims = list(set(range(coord.ndim)) - set(dims_to_collapse))
        untouched_shape = [coord.shape[dim] for dim in untouched_dims]
        new_shape = untouched_shape + [end_size]
        dims = untouched_dims + dims_to_collapse

        # Then reshape the data so that the dimensions being aggregated
        # over are grouped 'at the end' (i.e. axis=-1).
        unrolled_data = np.transpose(coord.points, dims).reshape(new_shape)

        new_points = kernel.aggregate(unrolled_data, axis=-1)
        new_coord = coord.copy(points=new_points)
        return new_coord

    @staticmethod
    def _calc_new_dims(coord_dims, dims_to_collapse):
        """
            Calculate the new dimensions for the coordinate.
        :param coord_dims: the original dimensions
        :param dims_to_collapse: The dimensions which are being collapsed over
        :return: The new coordinates which the coord will take on the collapsed cube
        """
        new_dims = []
        # For each original dimension subtract one for every collapsed coordinate which came before it.
        # TODO: There must be a cleaner way to do this...
        for d in coord_dims:
            new_d = d
            for dc in dims_to_collapse:
                if d > dc:
                    new_d -= 1
            # If the dimension is one being collapsed then we don't include it in the new dimensions
            if d not in dims_to_collapse:
                new_dims.append(new_d)

        return new_dims

    @staticmethod
    def _update_aux_factories(data, *args, **kwargs):
        from cis.utils import listify
        d_list = listify(data)
        for d in d_list:
            for factory in d.aux_factories:
                factory.update(*args, **kwargs)

    def _gridded_full_collapse(self, kernel):
        from copy import deepcopy
        from cis.exceptions import ClassNotFoundError
        ag_args = {}

        dims_to_collapse = set()
        for coord in self.coords:
            dims_to_collapse.update(self.data.coord_dims(coord))

        coords_for_partial_collapse = []

        # Collapse any coords that span the dimension(s) being collapsed
        for coord in self.data.aux_coords:
            coord_dims = self.data.coord_dims(coord)
            # If a coordinate has any of the dimensions we wan't to collapse AND has some dimensions we don't...
            if set(dims_to_collapse).intersection(coord_dims) and \
                    set(coord_dims).difference(dims_to_collapse):
                # ... add it to our list of partial coordinates to collapse.
                coords_for_partial_collapse.append((coord, coord_dims))

        if isinstance(kernel, iris.analysis.WeightedAggregator) and \
                        'latitude' in [c.standard_name for c in self.coords]:
            # If this is a list we can calculate weights using the first item (all variables should be on
            # same grid)
            data_for_weights = self.data[0] if isinstance(self.data, list) else self.data
            # Weights to correctly calculate areas.
            ag_args['weights'] = iris.analysis.cartography.area_weights(data_for_weights)
        elif not isinstance(kernel, iris.analysis.Aggregator):
            raise ClassNotFoundError('Error - unexpected aggregator type.')

        # Before we remove the coordinates which need to be partially collapsed we take a copy of the data. We need
        #  this so that the aggregation doesn't have any side effects on the input data. This is particularly important
        #  when using a MultiKernel for which this routine gets called multiple times.
        data_for_collapse = deepcopy(self.data)

        for coord, _ in coords_for_partial_collapse:
            data_for_collapse.remove_coord(coord)

        # Having set-up the collapse we can now just defer to the Cube.collapse method for much of the leg-work
        new_data = iris.cube.Cube.collapsed(data_for_collapse, self.coords, kernel, **ag_args)

        for coord, old_dims in coords_for_partial_collapse:
            collapsed_coord = GriddedCollapsor._partially_collapse_multidimensional_coord(coord, dims_to_collapse)
            new_dims = GriddedCollapsor._calc_new_dims(old_dims, dims_to_collapse)

            new_data.add_aux_coord(collapsed_coord, new_dims)
            # If the coordinate we had to collapse manually was a dependency in an aux factory (which is quite likely)
            #  then we need to put it back in and fix the factory, this will update any missing dependencies.
            self._update_aux_factories(new_data, None, collapsed_coord)

        return new_data

    def __call__(self, kernel):
        from cis.data_io.gridded_data import GriddedDataList
        from cis.aggregation.collapse_kernels import MultiKernel

        # Make sure all coordinate have bounds - important for weighting and aggregating
        # Only try and guess bounds on Dim Coords
        for coord in self.data.coords(dim_coords=True):
            if not coord.has_bounds() and len(coord.points) > 1:
                coord.guess_bounds()
                logging.warning("Creating guessed bounds as none exist in file")
                new_coord_number = self.data.coord_dims(coord)
                self.data.remove_coord(coord.name())
                self.data.add_dim_coord(coord, new_coord_number)

        output = GriddedDataList([])
        if isinstance(kernel, MultiKernel):
            for sub_kernel in kernel.sub_kernels:
                sub_kernel_out = self._gridded_full_collapse(sub_kernel)
                output.append_or_extend(sub_kernel_out)
        else:
            output.append_or_extend(self._gridded_full_collapse(kernel))
        return output




from netCDF4 import Dataset
import os
import unittest
from hamcrest import assert_that, greater_than_or_equal_to, less_than_or_equal_to, is_
from functools import reduce


class BaseIntegrationTest(unittest.TestCase):

    OUTPUT_FILENAME = "test_integration_out.nc"

    def setUp(self):
        # Set force overwrite in case working files are still present
        os.environ['CIS_FORCE_OVERWRITE'] = "True"
        self.clean_output()

    def tearDown(self):
        # Pop off the environemnt variable
        os.environ.pop('CIS_FORCE_OVERWRITE')
        self.clean_output()

    def clean_output(self):
        if hasattr(self, 'ds') and self.ds.isopen():
            self.ds.close()
        if os.path.exists(self.OUTPUT_FILENAME):
            os.remove(self.OUTPUT_FILENAME)

    def check_output_contains_variables(self, output_path, var_names):
        self.ds = Dataset(output_path)
        for var in var_names:
            try:
                var = self.ds.variables[var]
            except KeyError:
                raise AssertionError("Variable %s not found in output file" % var)
        self.ds.close()

    def check_output_file_variable_attribute_contains_string(self, output_path, variable, attribute, string):
        self.ds = Dataset(output_path)
        try:
            var = self.ds.variables[variable]
        except KeyError:
            raise AssertionError("Variable %s not found in output file" % variable)
        try:
            att_string = getattr(var, attribute)
        except AttributeError:
            raise AssertionError("Attribute %s not found in variable" % attribute)
        assert string in att_string
        self.ds.close()

    def check_latlon_subsetting(self, lat_max, lat_min, lon_max, lon_min, lat_var='latitude', lon_var='longitude'):
        self.ds = Dataset(self.OUTPUT_FILENAME)
        lat = self.ds.variables[lat_var][:]
        lon = self.ds.variables[lon_var][:]
        assert_that(min(lon), greater_than_or_equal_to(lon_min))
        assert_that(max(lon), less_than_or_equal_to(lon_max))
        assert_that(min(lat), greater_than_or_equal_to(lat_min))
        assert_that(max(lat), less_than_or_equal_to(lat_max))
        self.ds.close()

    def check_alt_subsetting(self, alt_max, alt_min):
        self.ds = Dataset(self.OUTPUT_FILENAME)
        alt = self.ds.variables['altitude'][:]
        assert_that(min(alt), greater_than_or_equal_to(alt_min))
        assert_that(max(alt), less_than_or_equal_to(alt_max))
        self.ds.close()

    def check_pres_subsetting(self, pres_max, pres_min, pres_name='air_pressure'):
        import numpy as np
        self.ds = Dataset(self.OUTPUT_FILENAME)
        pres = self.ds.variables[pres_name][:]
        assert_that(np.min(pres), greater_than_or_equal_to(pres_min))
        assert_that(np.max(pres), less_than_or_equal_to(pres_max))
        self.ds.close()

    @staticmethod
    def _clean_sample_file_name(sample_file):
        import re
        return re.sub(r'([\\]):', r':', sample_file)

    def check_output_col_grid(self, sample_file, sample_var, output_file, output_vars, expected_shape=None):
        """
        Check that the output grid matches the sample grid in shape.
        :param sample_file:
        :param sample_var:
        :param output_file:
        :param output_vars:
        :return:
        """
        from cis import read_data
        from operator import mul
        if expected_shape is None:
            sample_shape = read_data(self._clean_sample_file_name(sample_file), sample_var).data.shape
        else:
            sample_shape = expected_shape
        self.ds = Dataset(self._clean_sample_file_name(output_file))
        for output_var in output_vars:
            output_shape = self.ds.variables[output_var].shape
            # This copes with dims in different orders, length 1 values being taken out etc
            assert_that(reduce(mul, sample_shape), is_(reduce(mul, output_shape)))
        self.ds.close()

    def check_output_vars_are_different(self, output_file, output_vars):
        """
        Check that the output variables are NOT exactly the same
        :param output_file:
        :param output_vars:
        :return:
        """
        from itertools import combinations
        import numpy as np
        self.ds = Dataset(self._clean_sample_file_name(output_file))
        # Loop over each possible pair of output var
        for a, b in combinations(output_vars, 2):
            a_data = self.ds.variables[a]
            b_data = self.ds.variables[b]
            assert not np.allclose(a_data, b_data)
        self.ds.close()


from hamcrest import *
from nose.tools import istest, raises
import numpy
from cis.data_io.Coord import Coord, CoordList
from cis.data_io.ungridded_data import Metadata
from cis.exceptions import DuplicateCoordinateError


def create_dummy_coordinates_list():
    coord1 = Coord(numpy.array([5, 4]), Metadata(standard_name='grid_latitude'), axis='Y')
    coord2 = Coord(numpy.array([5, 4]), Metadata(standard_name='grid_longitude'), axis='X')
    return CoordList([coord1, coord2])


@istest
def can_create_a_valid_list_of_coordinates():
    list = create_dummy_coordinates_list()
    assert (len(list) == 2)
    assert (list[0].standard_name == 'grid_latitude')
    assert (list[1].standard_name == 'grid_longitude')
    assert (list[0].axis == 'Y')
    assert (list[1].axis == 'X')


@istest
def can_append_to_list_of_coordinates():
    list = create_dummy_coordinates_list()
    list.append(Coord(numpy.array([5, 4]), Metadata(standard_name='altitude'), axis='Z'))
    assert (len(list) == 3)
    assert (list[2].standard_name == 'altitude')
    assert (list[2].axis == 'Z')


@istest
@raises(DuplicateCoordinateError)
def append_a_duplicate_to_a_list_of_coordinates_fails():
    list = create_dummy_coordinates_list()
    list.append(Coord(numpy.array([5, 4]), Metadata(standard_name='grid_longitude'), axis='X'))


@istest
def can_find_a_coord_from_a_list_of_coordinates():
    list = create_dummy_coordinates_list()
    coord = list.get_coord(standard_name='grid_longitude')
    assert (coord.standard_name == 'grid_longitude')
    assert (coord.axis == 'X')


@istest
def can_find_many_coords_from_a_list_of_coordinates():
    list = create_dummy_coordinates_list()
    list.append(Coord(numpy.array([5, 4]), Metadata(name='testZ'), axis='Z'))
    list.append(Coord(numpy.array([5, 4]), Metadata(name='testZ', standard_name='air_pressure')))
    assert (len(list) == 4)
    coords = list.get_coords(var_name='testZ')
    assert (len(coords) == 2)
    assert (coords[0].axis == 'Z')
    assert (coords[1].axis == '')
    assert (coords[1].name() == 'air_pressure')

@istest
def can_convert_time_without_since_in_units():
    times = numpy.array([0, 1])
    units = "Days from the file reference point 1601-01-01"
    time_stamp_info = "1601-01-01"
    coord = Coord(times, Metadata(units=units))

    coord.convert_to_std_time(time_stamp_info)

    assert_that(coord.data_flattened[0], is_(366.0), "time point")


@istest
def can_convert_time_with_since_in_units():
    times = numpy.array([0, 1])
    units = "days since 1601-01-01"
    coord = Coord(times, Metadata(units=units))

    coord.convert_to_std_time()

    assert_that(coord.data_flattened[0], is_(366.0), "time point")


@istest
@raises(ValueError)
def can_not_convert_time_without_since_in_units_but_no_timestamp():
    times = numpy.array([0, 1])
    units = "Days from the file reference point 1601-01-01"
    coord = Coord(times, Metadata(units=units))

    coord.convert_to_std_time()


@istest
@raises(ValueError)
def can_not_convert_time_without_since_in_units_with_no_units():
    times = numpy.array([0, 1])
    units = ""
    time_stamp_info = "1601-01-01"
    coord = Coord(times, Metadata(units=units))

    coord.convert_to_std_time(time_stamp_info)



from nose.tools import istest, raises
from cis.data_io.products.AProduct import ProductPluginException
from cis.test.integration_test_data import cis_test_files, invalid_filename
from cis import read_data, read_data_list


@istest
@raises(ProductPluginException)
def test_read_data_raises_error_on_invalid_variable():
    file = cis_test_files['Aerosol_CCI'].master_filename
    read_data(file, 'invalid_variable')


@istest
@raises(ValueError)
def test_read_data_raises_error_on_more_than_one_variable_returned():
    file = cis_test_files['Aerosol_CCI'].master_filename
    read_data(file, cis_test_files['Aerosol_CCI'].all_variable_names)


@istest
@raises(IOError)
def test_read_data_raises_error_on_invalid_file():
    read_data(invalid_filename, cis_test_files['Aerosol_CCI'].all_variable_names)


@istest
@raises(IOError)
def test_read_data_raises_error_on_invalid_file_wildcard():
    read_data(invalid_filename + '*', cis_test_files['Aerosol_CCI'].all_variable_names)


@istest
@raises(ProductPluginException)
def test_read_data_list_raises_error_on_invalid_variable():
    file = cis_test_files['Aerosol_CCI'].master_filename
    read_data_list(file, 'invalid_variable')


@istest
@raises(IOError)
def test_read_data_list_raises_error_on_invalid_file():
    read_data_list(invalid_filename, cis_test_files['Aerosol_CCI'].all_variable_names)


@istest
@raises(IOError)
def test_read_data_list_raises_error_on_invalid_file_wildcard():
    read_data_list(invalid_filename + '*', cis_test_files['Aerosol_CCI'].all_variable_names)
